from .imx500 import IMX500, NetworkIntrinsics
from .postprocess_efficientdet_lite0 import \
    postprocess_efficientdet_lite0_detection
from .postprocess_nanodet import postprocess_nanodet_detection
from .postprocess_yolov5 import postprocess_yolov5_detection
from .postprocess_yolov8 import postprocess_yolov8_detection



from threading import Event

from .output import Output


class SplittableOutput(Output):
    """
    The SplittableOutput passes the encoded bitstream to another output (or drops them if there isn't one).

    It can be told to "split" the current output that it's been writing to, to another one. This means
    it switches seamlessly from the current output, which is closed, to a new one, without dropping
    any frames. By default, it performs the switch at a video keyframem though it can be told not to
    wait for one (by setting wait_for_keyframe to False).
    """

    def __init__(self, output=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._output = output
        self._new_output = None
        self._split_done = Event()
        self._streams = []

    def split_output(self, new_output, wait_for_keyframe=True):
        """
        Terminate the current output, switching seamlessly to the new one.

        If wait_for_keyframe is True, the switch only occurs when the next video keyframe is
        seen. Otherwise the switch happens immediately.

        THe function returns only when the switch has happened, which may entail waiting for a
        video keyframe, and the old output has been closed.
        """
        old_output = self._output
        # Start the new outoput in this thread, then schedule outputframe to make the switch.
        new_output.start()
        for encoder_stream, codec, kwargs in self._streams:
            new_output._add_stream(encoder_stream, codec, **kwargs)
        self._wait_for_keyframe = wait_for_keyframe
        self._new_output = new_output
        # Wait for the switch-over to happen, and close the old output in this thread too.
        self._split_done.wait()
        self._split_done.clear()
        if old_output:
            old_output.stop()

    def outputframe(self, frame, keyframe=True, timestamp=None, packet=None, audio=False):
        # Audio frames probably always say they're keyframes, but we must wait for a video one.
        if self._new_output and (not self._wait_for_keyframe or (not audio and keyframe)):
            self._split_done.set()
            # split_output will close the old output.
            self._output = self._new_output
            self._new_output = None
        if self._output:
            self._output.outputframe(frame, keyframe, timestamp, packet, audio)

    def start(self):
        super().start()
        if self._output:
            self._output.start()

    def stop(self):
        super().stop()
        if self._output:
            self._output.stop()

    def _add_stream(self, encoder_stream, codec_name, **kwargs):
        # The underlying output may need to know what streams it's dealing with, so we must
        # remember them.
        self._streams.append((encoder_stream, codec_name, kwargs))
        # Forward immediately to the output if we were given one initially.
        if self._output:
            self._output._add_stream(encoder_stream, codec_name, **kwargs)



"""
Highernet postprocessing

This code is based on multiple sources:
https://github.com/HRNet/HigherHRNet-Human-Pose-Estimation
https://github.com/princeton-vl/pose-ae-train
https://github.com/yinguobing/facial-landmark-detection-hrnet
"""

from typing import Tuple

import cv2
import numpy as np

try:
    from munkres import Munkres
except ImportError:
    raise ImportError("Please install munkres first. `pip3 install --break-system-packages munkres`")

default_joint_order = [0, 1, 2, 3, 4, 5, 6, 11, 12, 7, 8, 9, 10, 13, 14, 15, 16]


def postprocess_higherhrnet(outputs: list[np.ndarray, np.ndarray],
                            img_size,
                            img_w_pad,
                            img_h_pad,
                            network_postprocess,
                            num_joints=17,
                            tag_per_joint=True,
                            joint_order=default_joint_order,
                            detection_threshold=0.3,
                            max_num_people=30,
                            nms_kernel=5,
                            nms_padding=2,
                            ignore_too_much=False,
                            use_detection_val=True,
                            tag_threshold=1.0,
                            adjust=False,
                            refine=False,
                            input_image_size=(288, 384),
                            output_shape=(144, 192)) -> Tuple[list[list], list, list[list]]:
    all_preds = []
    all_scores = []
    if network_postprocess:
        # outputs [[B, max_num_people, num_joints], [B, max_num_people, num_joints], [B, max_num_people, num_joints]]
        grouped, scores = parse(network_outputs=[outputs[0][0, ...],
                                                 outputs[1][0, ...],
                                                 outputs[2][0, ...]],
                                output_shape=output_shape,
                                adjust=adjust,
                                refine=refine,
                                network_postprocess=network_postprocess,
                                tag_per_joint=tag_per_joint,
                                max_num_people=max_num_people,
                                nms_kernel=nms_kernel,
                                nms_padding=nms_padding,
                                num_joints=num_joints,
                                joint_order=joint_order,
                                detection_threshold=detection_threshold,
                                ignore_too_much=ignore_too_much,
                                use_detection_val=use_detection_val,
                                tag_threshold=tag_threshold)
    else:
        out0 = outputs[0][0]
        out1 = outputs[1][0]

        # postprocess:
        # resize first output to 2nd output size
        out0 = ResizeBilinear(out0, out1.shape[0], out1.shape[1])
        # average heatmaps from both outputs
        heatmaps = (out0[..., :17] + out1) / 2
        tags = out0[..., 17:]
        grouped, scores = parse(network_outputs=[heatmaps, tags],
                                output_shape=output_shape,
                                adjust=adjust,
                                refine=refine,
                                network_postprocess=network_postprocess,
                                tag_per_joint=tag_per_joint,
                                max_num_people=max_num_people,
                                nms_kernel=nms_kernel,
                                nms_padding=nms_padding,
                                num_joints=num_joints,
                                joint_order=joint_order,
                                detection_threshold=detection_threshold,
                                ignore_too_much=ignore_too_much,
                                use_detection_val=use_detection_val,
                                tag_threshold=tag_threshold)

    # scale keypoints coordinates to input image size
    scale_factor = (np.array(input_image_size) / output_shape).reshape((1, 1, 2))
    for img_index in range(len(grouped)):
        if grouped[img_index].shape[0] > 0:
            # rescale to preprocessed input image size
            grouped[img_index][:, :, :2] = grouped[img_index][:, :, :2] * scale_factor
            # remove pad offset:
            grouped[img_index][:, :, 0] = grouped[img_index][:, :, 0] - img_w_pad[0]
            grouped[img_index][:, :, 1] = grouped[img_index][:, :, 1] - img_h_pad[0]
            # rescale to original image size
            resized_input_image = np.array(input_image_size) - np.array(
                (sum(img_h_pad),
                 sum(img_w_pad)))
            s = (np.array(img_size) / resized_input_image).reshape((1, 1, 2))
            grouped[img_index][:, :, :2] = grouped[img_index][:, :, :2] * s

    # Calculate zero keypoint
    zero_kpt = np.zeros((1, 4))
    resized_input_image = np.array(input_image_size) - np.array(
        (sum(img_h_pad),
         sum(img_w_pad)))
    s = (np.array(img_size) / resized_input_image).reshape((1, 1, 2))
    zero_kpt[:, 0] = zero_kpt[:, 0] - img_w_pad[0]
    zero_kpt[:, 1] = zero_kpt[:, 1] - img_h_pad[0]
    zero_kpt[:, :2] = zero_kpt[:, :2] * s

    all_preds.append(grouped)
    all_scores.append(scores)

    kpts = []
    # one image, one iter
    for idx, _kpts in enumerate(all_preds):
        for idx_kpt, kpt in enumerate(_kpts[0]):
            area = (np.max(kpt[:, 0]) - np.min(kpt[:, 0])) * (np.max(kpt[:, 1]) - np.min(kpt[:, 1]))
            # kpt [17, 4]
            kpt = processKeypoints(kpt)
            kpts.append(
                {
                    'keypoints': kpt[:, 0:3],
                    'score': all_scores[idx][idx_kpt],
                    'tags': kpt[:, 3],
                    'area': area
                }
            )
    # _coco_keypoint_results_one_category_kernel
    out_keypoints = []
    out_scores = []
    out_bbox = []

    # for img_kpts in kpts:
    img_kpts = kpts
    if len(img_kpts) == 0:
        return [], [], []

    _key_points = np.array(
        [img_kpts[k]['keypoints'] for k in range(len(img_kpts))]
    )
    key_points = np.zeros(
        (_key_points.shape[0], num_joints * 3),
        dtype=np.float32
    )

    for ipt in range(num_joints):
        key_points[:, ipt * 3 + 0] = _key_points[:, ipt, 0]
        key_points[:, ipt * 3 + 1] = _key_points[:, ipt, 1]
        key_points[:, ipt * 3 + 2] = _key_points[:, ipt, 2]  # keypoints score.

    for k in range(len(img_kpts)):
        kpt = key_points[k].reshape((num_joints, 3))
        # ignore zero kpts
        mask = np.isin(kpt, zero_kpt)
        kpt = np.where(mask, np.nan, kpt)
        left_top = np.nanmin(kpt, axis=0)
        right_bottom = np.nanmax(kpt, axis=0)

        out_keypoints.append(list(key_points[k]))
        out_scores.append(img_kpts[k]['score'])
        out_bbox.append([left_top[1], left_top[0], right_bottom[1], right_bottom[0]])
    return out_keypoints, out_scores, out_bbox


def parse(network_outputs,
          output_shape,
          adjust=False,
          refine=False,
          network_postprocess=False,
          tag_per_joint=17,
          max_num_people=30,
          nms_kernel=5,
          nms_padding=2,
          num_joints=17,
          joint_order=default_joint_order,
          detection_threshold=0.1,
          ignore_too_much=False,
          use_detection_val=True,
          tag_threshold=1.0
          ):
    if network_postprocess:
        tag_k, ind_k, val_k = network_outputs
        x = ind_k % output_shape[1]
        y = (ind_k / output_shape[1]).astype(ind_k.dtype)
        ind_k = np.stack([x, y], axis=2)

        topk_output_dict = {'tag_k': tag_k[np.newaxis, ...],
                            'loc_k': ind_k[np.newaxis, ...],
                            'val_k': val_k[np.newaxis, ...],
                            }
    else:
        det, tag = network_outputs
        # topk_output_dict
        # {'tag_k': [num_images, max_num_people, num_joints],
        # 'loc_k': [num_images, max_num_people, num_joints, 2],
        # 'val_k': [num_images, max_num_people, num_joints]}
        topk_output_dict = top_k(det=det,
                                 tag=tag,
                                 tag_per_joint=tag_per_joint,
                                 max_num_people=max_num_people,
                                 nms_kernel=nms_kernel,
                                 nms_padding=nms_padding)
    # ans [num_joints_detected, num_joints, 4]
    ans = match(tag_k=topk_output_dict['tag_k'],
                loc_k=topk_output_dict['loc_k'],
                val_k=topk_output_dict['val_k'],
                num_joints=num_joints,
                joint_order=joint_order,
                detection_threshold=detection_threshold,
                max_num_people=max_num_people,
                ignore_too_much=ignore_too_much,
                use_detection_val=use_detection_val,
                tag_threshold=tag_threshold)
    if adjust:
        # ans [[num_joints_detected, num_joints, 4]]
        ans = adjust_func(ans, det[np.newaxis, ...])  # TODO support batch size > 1

    scores = [i[:, 2].mean() for i in ans[0]]

    if refine:
        ans = ans[0]
        # for every detected person
        for _ in range(len(ans)):
            # NotImplemented
            if not tag_per_joint:
                raise NotImplementedError

        # ans [[num_joints_detected, num_joints, 4]]
        ans = [ans]
    return ans, scores


def ResizeBilinear(img, new_height, new_width):
    return cv2.resize(img, (new_width, new_height))


def top_k(det,
          tag,
          tag_per_joint=17,
          max_num_people=30,
          nms_kernel=5,
          nms_padding=2):
    # det [144, 192, 17]
    # tag [144, 192, 17]

    # det [144, 192, 17]
    det = nms(det,
              nms_kernel=nms_kernel,
              nms_padding=nms_padding)
    # num_images 1
    # h 144
    # w 192
    # num_joints 17
    num_images, h, w, num_joints = (1,) + det.shape  # TODO: support multiple images (batch>1)

    # det [num_images, h*w, num_joints]
    det = det.reshape((num_images, -1, num_joints))
    # val_k [num_images, max_num_people, num_joints]
    val_k, ind = np_topk(det, max_num_people)

    # tag [num_images, h*w, num_joints]
    tag = tag.reshape((num_images, -1, num_joints))

    # NotImplemented
    if not tag_per_joint:
        raise NotImplementedError
        tag = tag.expand(-1, num_joints, -1, -1)

    # tag_k [num_images, max_num_people, num_joints]
    tag_k = np.zeros((num_images, max_num_people, num_joints))
    for img in range(num_images):
        for kp in range(num_joints):
            tag_k[img, :, kp] = tag[img, ind[img, :, kp], kp]

    x = ind % w
    y = (ind / w).astype(ind.dtype)

    # ind_k [num_images, max_num_people, num_joints, 2]
    ind_k = np.stack([x, y], axis=3)

    # {'tag_k': [num_images, max_num_people, num_joints],
    # 'loc_k': [num_images, max_num_people, num_joints, 2],
    # 'val_k': [num_images, max_num_people, num_joints]}
    return {'tag_k': tag_k,
            'loc_k': ind_k,
            'val_k': val_k,
            }


def nms(det,
        nms_kernel=5,
        nms_padding=2):
    # det [144, 192, 17]
    # maxm [144, 192, 17]
    maxm = np_max_pool(det, k=nms_kernel, p=nms_padding)
    maxm = np.equal(maxm, det).astype(np.float32)
    det = det * maxm
    return det


def np_max_pool(x,
                k=5,
                p=2,
                p_value=0):
    # x [144, 192, 17]
    # k - kernel size (h, w)
    # p - padding size (top, bottom, left, right)
    if isinstance(k, int):
        k = (k, k)
    if isinstance(p, int):
        p = ((p, p), (p, p), (0, 0))
    elif isinstance(p, (list, tuple)) and len(p) == 2:
        p = ((p[0], p[0]), (p[1], p[1]), (0, 0))

    # y [148, 196, 17
    y = np.pad(x, p)
    out = np.concatenate(
        [np.max(np.concatenate([y[ky:ky + y.shape[0] - k[0] + 1, kx:kx + y.shape[1] - k[1] + 1, c:c + 1]
                                for ky in range(k[0])
                                for kx in range(k[1])], 2), axis=2, keepdims=True) for c in range(y.shape[2])], 2)
    # out [144, 192, 17]
    return out


def np_topk(x, k):
    # x [1, 27648, 17]
    # n_images 1
    # n_keypoints 17
    n_images, _, n_keypoints = x.shape
    # vals [1, k, 17]
    # inds [1, k, 17]
    vals = np.zeros((n_images, k, n_keypoints), dtype=x.dtype)
    inds = np.zeros((n_images, k, n_keypoints), dtype=np.int64)
    for img in range(n_images):
        for kp in range(n_keypoints):
            # _inds [k]
            _inds = np.argpartition(x[img, :, kp], -k)[-k:]
            _inds = _inds[np.argsort(x[img, _inds, kp], )][::-1]
            inds[img, :, kp] = _inds
            vals[img, :, kp] = x[img, _inds, kp]
    return vals, inds


def match(tag_k,
          loc_k,
          val_k,
          num_joints=17,
          joint_order=default_joint_order,
          detection_threshold=0.1,
          max_num_people=30,
          ignore_too_much=False,
          use_detection_val=True,
          tag_threshold=1.0):
    def m(x):
        return match_by_tag(inp=x,
                            num_joints=num_joints,
                            joint_order=joint_order,
                            detection_threshold=detection_threshold,
                            max_num_people=max_num_people,
                            ignore_too_much=ignore_too_much,
                            use_detection_val=use_detection_val,
                            tag_threshold=tag_threshold)
    return list(map(m, zip(tag_k, loc_k, val_k)))


def match_by_tag(inp,
                 num_joints=17,
                 joint_order=default_joint_order,
                 detection_threshold=0.1,
                 max_num_people=30,
                 ignore_too_much=False,
                 use_detection_val=True,
                 tag_threshold=1.0):
    # tag_k [num_images, max_num_people, num_joints]
    # loc_k [num_images, max_num_people, num_joints, 2]
    # val_k [num_images, max_num_people, num_joints]
    tag_k, loc_k, val_k = inp
    # default_ [num_joints, 4]
    default_ = np.zeros((num_joints, 3 + 1))  # tag_k.shape[2] assumed to be 1  # pytorch shape: (17, 4)

    joint_dict = {}
    tag_dict = {}
    for i in range(num_joints):
        idx = joint_order[i]

        # tags [max_num_people, 1]
        tags = tag_k[:, idx:idx + 1]
        # joints [max_num_people, 4]
        joints = np.concatenate((loc_k[:, idx, :], val_k[:, idx:idx + 1], tags), 1)
        # mask [max_num_people]
        mask = joints[:, 2] > detection_threshold
        tags = tags[mask]
        joints = joints[mask]

        if joints.shape[0] == 0:
            continue

        if i == 0 or len(joint_dict) == 0:
            for tag, joint in zip(tags, joints):
                key = tag[0]
                joint_dict.setdefault(key, np.copy(default_))[idx] = joint
                tag_dict[key] = [tag]
        else:
            grouped_keys = list(joint_dict.keys())[:max_num_people]
            grouped_tags = [np.mean(tag_dict[i], axis=0) for i in grouped_keys]

            if ignore_too_much \
                    and len(grouped_keys) == max_num_people:
                continue

            diff = joints[:, None, 3:] - np.array(grouped_tags)[None, :, :]
            diff_normed = np.linalg.norm(diff, ord=2, axis=2)
            diff_saved = np.copy(diff_normed)

            if use_detection_val:
                diff_normed = np.round(diff_normed) * 100 - joints[:, 2:3]

            num_added = diff.shape[0]
            num_grouped = diff.shape[1]

            if num_added > num_grouped:
                diff_normed = np.concatenate(
                    (
                        diff_normed,
                        np.zeros((num_added, num_added - num_grouped)) + 1e10
                    ),
                    axis=1
                )

            pairs = py_max_match(diff_normed)
            for row, col in pairs:
                if (
                        row < num_added
                        and col < num_grouped
                        and diff_saved[row][col] < tag_threshold
                ):
                    key = grouped_keys[col]
                    joint_dict[key][idx] = joints[row]
                    tag_dict[key].append(tags[row])
                else:
                    key = tags[row][0]
                    joint_dict.setdefault(key, np.copy(default_))[idx] = \
                        joints[row]
                    tag_dict[key] = [tags[row]]

    # ans [len(joint_dict), num_joints, 4]
    ans = np.array([joint_dict[i] for i in joint_dict]).astype(np.float32)
    return ans


def py_max_match(scores):
    m = Munkres()
    tmp = m.compute(scores)
    tmp = np.array(tmp).astype(np.int32)
    return tmp


def adjust_func(ans, det):
    # ans [[num_joints_detected, num_joints, 4]]
    # det [144, 192, 17]
    for batch_id, people in enumerate(ans):
        for people_id, i in enumerate(people):
            for joint_id, joint in enumerate(i):
                if joint[2] > 0:
                    y, x = joint[0:2]
                    xx, yy = int(x), int(y)
                    # print(batch_id, joint_id, det[batch_id].shape)
                    tmp = det[batch_id][..., joint_id]
                    if tmp[xx, min(yy + 1, tmp.shape[1] - 1)] > tmp[xx, max(yy - 1, 0)]:
                        y += 0.25
                    else:
                        y -= 0.25

                    if tmp[min(xx + 1, tmp.shape[0] - 1), yy] > tmp[max(0, xx - 1), yy]:
                        x += 0.25
                    else:
                        x -= 0.25
                    ans[batch_id][people_id, joint_id, 0:2] = (y + 0.5, x + 0.5)
    # ans [[num_joints_detected, num_joints, 4]]
    return ans


def refine_func(det, tag, keypoints):
    # det [144, 192, 17]
    # tag [144, 192, 17]
    # keypoints [num_joints, 4]
    if len(tag.shape) == 3:
        # tag shape: (17, 128, 128, 1)
        # tag [144, 192, 17, 1]
        tag = tag[:, :, :, None]

    tags = []
    for i in range(keypoints.shape[0]):
        if keypoints[i, 2] > 0:
            # save tag value of detected keypoint
            x, y = keypoints[i][:2].astype(np.int32)
            tags.append(tag[y, x, i])

    # mean tag of current detected people
    prev_tag = np.mean(tags, axis=0)
    ans = []

    for i in range(keypoints.shape[0]):
        # score of joints i at all position
        tmp = det[:, :, i]
        # distance of all tag values with mean tag of current detected people
        tt = (((tag[:, :, i] - prev_tag[None, None, :]) ** 2).sum(axis=2) ** 0.5)
        tmp2 = tmp - np.round(tt)

        # find maximum position
        y, x = np.unravel_index(np.argmax(tmp2), tmp.shape)
        xx = x
        yy = y
        # detection score at maximum position
        val = tmp[y, x]
        # offset by 0.5
        x += 0.5
        y += 0.5

        # add a quarter offset
        if tmp[yy, min(xx + 1, tmp.shape[1] - 1)] > tmp[yy, max(xx - 1, 0)]:
            x += 0.25
        else:
            x -= 0.25

        if tmp[min(yy + 1, tmp.shape[0] - 1), xx] > tmp[max(0, yy - 1), xx]:
            y += 0.25
        else:
            y -= 0.25

        ans.append((x, y, val))
    ans = np.array(ans)

    if ans is not None:
        for i in range(det.shape[2]):
            # add keypoint if it is not detected
            if ans[i, 2] > 0 and keypoints[i, 2] == 0:
                # if ans[i, 2] > 0.01 and keypoints[i, 2] == 0:
                keypoints[i, :2] = ans[i, :2]
                keypoints[i, 2] = ans[i, 2]
    # keypoints [num_joints_detected, num_joints, 4]
    return keypoints


def processKeypoints(keypoints):
    # keypoints [17, 4]
    tmp = keypoints.copy()
    if keypoints[:, 2].max() > 0:
        num_keypoints = keypoints.shape[0]
        for i in range(num_keypoints):
            tmp[i][0:3] = [
                float(keypoints[i][0]),
                float(keypoints[i][1]),
                float(keypoints[i][2])
            ]

    return tmp



"""Encoder functionality"""

import threading
from enum import Enum

import av
from libcamera import controls

import picamera2.formats as formats

from ..outputs import Output
from ..request import _MappedBuffer


class Quality(Enum):
    """Enum type to describe the quality wanted from an encoder.

    This may be passed if a specific value (such as bitrate) has not been set.
    """

    VERY_LOW = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3
    VERY_HIGH = 4


class Encoder:
    """
    Base class for encoders.

    Mostly this defines the API for derived encoder classes, but it also handles optional audio encoding.
    For audio, a separate thread is started, which encodes audio packets and forwards them to the
    encoder's output object(s). This only work when the output object understands the audio stream,
    meaning that (at the time of writing) this must be a PyavOutput (though you could send output there
    via a CircularOutput2).

    Additional audio parameters:
    audio - set to True to enable audio encoding and output.
    audio_input - list of parameters that is passed to PyAv.open to create the audio input.
    audio_output - list of parameters passed to PyAv add_stream to define the audio codec and output stream.
    audio_sync - value (in us) by which to advance the audio stream to better sync with the video.

    Reasonable defaults are supplied so that applications can often just set the audio property to True.
    The audio_input and audio_output parameters are passed directly to PyAV, so will accept whatever PyAV
    understands.
    """

    def __init__(self):
        """Initialises encoder"""
        self._width = 0
        self._height = 0
        self._stride = 0
        self._format = None
        self._output = []
        self._running = False
        self._name = None
        self._lock = threading.Lock()
        self.firsttimestamp = None
        self.frame_skip_count = 1
        self._skip_count = 0
        self._output_lock = threading.Lock()
        # Set to True to enable audio.
        self.audio = False
        # These parameters are passed to Pyav to open the input audio container.
        self.audio_input = {'file': 'default', 'format': 'pulse'}
        # THese parameters are passed to Pyav for creating the encoded audio output stream.
        self.audio_output = {'codec_name': 'aac'}
        self.audio_sync = -100000  # in us, so by default, delay audio by 100ms
        self._audio_start = threading.Event()
        self.frames_encoded = 0
        # For camera sync.
        self.sync_enable = False
        self.sync = threading.Event()

    @property
    def running(self):
        return self._running

    @property
    def width(self):
        """Gets width

        :return: Width of frames
        :rtype: int
        """
        return self._width

    @width.setter
    def width(self, value):
        """Sets width

        :param value: Width
        :type value: int
        :raises RuntimeError: Failed to set width
        """
        if not isinstance(value, int):
            raise RuntimeError("Width must be integer")
        self._width = value

    @property
    def height(self):
        """Gets height

        :return: Height of frames
        :rtype: int
        """
        return self._height

    @height.setter
    def height(self, value):
        """Sets height

        :param value: Height
        :type value: int
        :raises RuntimeError: Failed to set height
        """
        if not isinstance(value, int):
            raise RuntimeError("Height must be integer")
        self._height = value

    @property
    def size(self):
        """Gets size

        :return: Size of frames as (width, height)
        :rtype: tuple
        """
        return (self._width, self._height)

    @size.setter
    def size(self, value):
        """Sets size

        :param value: Size
        :type value: tuple
        :raises RuntimeError: Failed to set size
        """
        if not isinstance(value, tuple) or len(value) != 2:
            raise RuntimeError("Size must be a tuple of two integers")
        self.width, self.height = value

    @property
    def stride(self):
        """Gets stride

        :return: Stride
        :rtype: int
        """
        return self._stride

    @stride.setter
    def stride(self, value):
        """Sets stride

        :param value: Stride
        :type value: int
        :raises RuntimeError: Failed to set stride
        """
        if not isinstance(value, int):
            raise RuntimeError("Stride must be integer")
        self._stride = value

    @property
    def format(self):
        """Get current format

        :return: Current format
        :rtype: int
        """
        return self._format

    @format.setter
    def format(self, value):
        """Sets input format to encoder

        :param value: Format
        :type value: str
        :raises RuntimeError: Invalid format
        """
        formats.assert_format_valid(value)
        self._format = value

    @property
    def output(self):
        """Gets output objects

        :return: Output object list or single Output object
        :rtype: List[Output]
        """
        if len(self._output) == 1:
            return self._output[0]
        else:
            return self._output

    @output.setter
    def output(self, value):
        """Sets output object, to write frames to

        :param value: Output object
        :type value: Output
        :raises RuntimeError: Invalid output passed
        """
        if isinstance(value, list):
            for out in value:
                if not isinstance(out, Output):
                    raise RuntimeError("Must pass Output")
        elif isinstance(value, Output):
            value = [value]
        else:
            raise RuntimeError("Must pass Output")
        self._output = value

    @property
    def name(self):
        """Gets stream name

        :return: Name
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value):
        """Sets stream name

        :param value: Name
        :type value: str
        :raises RuntimeError: Failed to set name
        """
        if not isinstance(value, str):
            raise RuntimeError("Name must be string")
        self._name = value

    def encode(self, stream, request):
        """Encode a frame

        :param stream: Stream
        :type stream: stream
        :param request: Request
        :type request: request
        """
        if self.audio:
            self._audio_start.set()  # Signal the audio encode thread to start.

        # If "sync" has been requested, we must wait for the image metadata to say that we
        # don't need to wait any more. While waiting, we simply don't encode any frames.
        if self.sync_enable:
            metadata = request.get_metadata()
            if metadata.get('SyncReady', False):
                self.sync_enable = False
                self.sync.set()
            else:
                return

        if self._skip_count == 0:
            with self._lock:
                if not self._running:
                    return
                self._encode(stream, request)
        self._skip_count = (self._skip_count + 1) % self.frame_skip_count
        self.frames_encoded += 1

    def _encode(self, stream, request):
        if isinstance(stream, str):
            stream = request.stream_map[stream]
        timestamp_us = self._timestamp(request)
        with _MappedBuffer(request, stream) as b:
            self.outputframe(b, keyframe=True, timestamp=timestamp_us)

    def start(self, quality=None):
        with self._lock:
            if self._running:
                raise RuntimeError("Encoder already running")
            self.frames_encoded = 0
            self._setup(quality)
            self._running = True
            self.firsttimestamp = None
            for out in self._output:
                out.start()
            self._start()

            # Start the audio, if that's been requested.
            if self.audio:
                self._audio_input_container = av.open(**self.audio_input)
                self._audio_input_stream = self._audio_input_container.streams.get(audio=0)[0]
                self._audio_output_container = av.open("/dev/null", 'w', format="null")
                self._audio_output_stream = self._audio_output_container.add_stream(**self.audio_output)
                # Outputs that can handle audio need to be told about its existence.
                for out in self._output:
                    out._add_stream(self._audio_output_stream, **self.audio_output)
                self._audio_thread = threading.Thread(target=self._audio_thread_func, daemon=True)
                self._audio_start.clear()
                self._audio_thread.start()  # audio thread will wait for the _audio_start event.

    def _start(self):
        pass

    def stop(self):
        with self._lock:
            if not self._running:
                raise RuntimeError("Encoder already stopped")
            self._running = False
            self._stop()
            if self.audio:
                self._audio_start.set()  # just in case it wasn't!
                self._audio_thread.join()
                self._audio_input_container.close()
                self._audio_output_container.close()
            for out in self._output:
                out.stop()

    def _stop(self):
        pass

    def outputframe(self, frame, keyframe=True, timestamp=None, packet=None, audio=False):
        """Writes a frame

        :param frame: Frame
        :type frame: bytes
        :param keyframe: Whether frame is a keyframe or not, defaults to True
        :type keyframe: bool, optional
        """
        with self._output_lock:
            for out in self._output:
                out.outputframe(frame, keyframe, timestamp, packet, audio)

    def _setup(self, quality):
        pass

    def _timestamp(self, request):
        # The sensor timestamp is the most accurate one, so we'll fetch that.
        ts = int(request.request.metadata[controls.SensorTimestamp] / 1000)  # ns to us
        if self.firsttimestamp is None:
            self.firsttimestamp = ts
            timestamp_us = 0
        else:
            timestamp_us = ts - self.firsttimestamp
        return timestamp_us

    def _handle_audio_packet(self, audio_packet):
        # Write out audio an packet, dealing with timestamp adjustments.
        time_scale_factor = 1000000 * self._audio_output_stream.codec_context.time_base
        delta = int(self.audio_sync / time_scale_factor)  # convert to audio time base
        audio_packet.pts -= delta
        audio_packet.dts -= delta
        timestamp = int(audio_packet.pts * time_scale_factor)  # want this in us
        if audio_packet.pts >= 0:
            self.outputframe(None, True, timestamp, audio_packet, True)

    def _audio_thread_func(self):
        # Audio thread that fetches audio packets, encodes them and forwards them to the output.
        # The output has to be able to understand audio, which means using a PyavOutput.
        # _audio_start gets signalled when the first video frame is submitted for encode, which will hopefully
        # keep the audio_sync adjustment more similar across different devices. Until that happens, though,
        # we must keep consuming and discarding the audio.
        for _ in self._audio_input_container.decode(self._audio_input_stream):
            if self._audio_start.isSet():
                break

        for audio_frame in self._audio_input_container.decode(self._audio_input_stream):
            if not self._running:
                break
            for audio_packet in self._audio_output_stream.encode(audio_frame):
                self._handle_audio_packet(audio_packet)

        # Flush out any remaining audio packets.
        for audio_packet in self._audio_output_stream.encode(None):
            self._handle_audio_packet(audio_packet)



from __future__ import annotations

import io
import logging
import time
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import libcamera
import numpy as np
import piexif
import simplejpeg
from pidng.camdefs import Picamera2Camera
from pidng.core import PICAM2DNG
from PIL import Image

import picamera2.formats as formats

from .controls import Controls
from .sensor_format import SensorFormat
from .utils import convert_from_libcamera_type

if TYPE_CHECKING:
    from picamera2.picamera2 import Picamera2

_log = logging.getLogger(__name__)


class _MappedBuffer:
    def __init__(self, request: "CompletedRequest", stream: str, write: bool = True) -> None:
        if isinstance(stream, str):
            stream = request.stream_map[stream]
        assert request.request is not None
        self.__fb = request.request.buffers[stream]
        self.__sync = request.picam2.allocator.sync(request.picam2.allocator, self.__fb, write)

    def __enter__(self) -> Any:
        self.__mm = self.__sync.__enter__()
        return self.__mm

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        self.__sync.__exit__(exc_type, exc_value, exc_traceback)


class MappedArray:
    def __init__(self, request: "CompletedRequest", stream: str, reshape: bool = True, write: bool = True) -> None:
        self.__request: "CompletedRequest" = request
        self.__stream: str = stream
        self.__buffer: _MappedBuffer = _MappedBuffer(request, stream, write=write)
        self.__array: Optional[np.ndarray] = None
        self.__reshape: bool = reshape

    def __enter__(self) -> "MappedArray":
        b = self.__buffer.__enter__()
        array = np.array(b, copy=False, dtype=np.uint8)

        if self.__reshape:
            if isinstance(self.__stream, str):
                config = self.__request.config[self.__stream]
            else:
                config = self.__stream.configuration

            # helpers._make_array_shared never makes a copy.
            array = self.__request.picam2.helpers._make_array_shared(array, config)

        self.__array = array
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        if self.__array is not None:
            del self.__array
        self.__buffer.__exit__(exc_type, exc_value, exc_traceback)

    @property
    def array(self) -> Optional[np.ndarray]:
        return self.__array


class CompletedRequest:
    FASTER_JPEG = True  # set to False to use the older JPEG encode method

    def __init__(self, request: Any, picam2: "Picamera2") -> None:
        self.request = request
        self.ref_count: int = 1
        self.lock = picam2.request_lock
        self.picam2 = picam2
        self.stop_count: int = picam2.stop_count
        self.configure_count: int = picam2.configure_count
        self.config = self.picam2.camera_config.copy()
        self.stream_map = self.picam2.stream_map.copy()
        with self.lock:
            self.syncs = [picam2.allocator.sync(self.picam2.allocator, buffer, False)
                          for buffer in self.request.buffers.values()]
            self.picam2.allocator.acquire(self.request.buffers)
            [sync.__enter__() for sync in self.syncs]

    def acquire(self) -> None:
        """Acquire a reference to this completed request, which stops it being recycled back to the camera system."""
        with self.lock:
            if self.ref_count == 0:
                raise RuntimeError("CompletedRequest: acquiring lock with ref_count 0")
            self.ref_count += 1

    def release(self) -> None:
        """Release this completed frame back to the camera system (once its reference count reaches zero)."""
        with self.lock:
            self.ref_count -= 1
            if self.ref_count < 0:
                raise RuntimeError("CompletedRequest: lock now has negative ref_count")
            elif self.ref_count == 0:
                # If the camera has been stopped since this request was returned then we
                # can't recycle it.
                if self.picam2.camera and self.stop_count == self.picam2.stop_count and self.picam2.started:
                    assert self.request is not None
                    self.request.reuse()
                    controls = self.picam2.controls.get_libcamera_controls()
                    for id, value in controls.items():

                        # libcamera now has "ExposureTimeMode" and "AnalogueGainMode" which must be set to
                        # manual for the fixed exposure time or gain to have any effect, and cleared to return
                        # to "auto " mode. We're going to hide that by supplying them automatically as needed.
                        if id == libcamera.controls.ExposureTime:
                            if value:
                                self.request.set_control(libcamera.controls.ExposureTimeMode, 1)  # manual
                            else:
                                self.request.set_control(libcamera.controls.ExposureTimeMode, 0)  # auto
                                continue  # no need to set the zero value!
                        elif id == libcamera.controls.AnalogueGain:
                            if value:
                                self.request.set_control(libcamera.controls.AnalogueGainMode, 1)  # manual
                            else:
                                self.request.set_control(libcamera.controls.AnalogueGainMode, 0)  # auto
                                continue  # no need to set the zero value!

                        self.request.set_control(id, value)

                    self.picam2.controls = Controls(self.picam2)
                    self.picam2.camera.queue_request(self.request)
                [sync.__exit__() for sync in self.syncs]
                assert self.request is not None
                self.picam2.allocator.release(self.request.buffers)
                self.request = None
                self.config = {}
                self.stream_map = {}

    def make_buffer(self, name: str) -> np.ndarray:
        """Make a 1D numpy array from the named stream's buffer."""
        if self.stream_map.get(name) is None:
            raise RuntimeError(f'Stream {name!r} is not defined')
        with _MappedBuffer(self, name, write=False) as b:
            return np.array(b, dtype=np.uint8)

    def get_metadata(self) -> Dict[str, Any]:
        """Fetch the metadata corresponding to this completed request."""
        metadata = {}
        assert self.request is not None
        for k, v in self.request.metadata.items():
            metadata[k.name] = convert_from_libcamera_type(v)
        return metadata

    def make_array(self, name: str) -> np.ndarray:
        """Make a 2d numpy array from the named stream's buffer."""
        config = self.config.get(name, None)
        if config is None:
            raise RuntimeError(f'Stream {name!r} is not defined')
        elif config['format'] == 'MJPEG':
            return np.array(Image.open(io.BytesIO(self.make_buffer(name))))

        # We don't want to send out an exported handle to the camera buffer, so we're going to have
        # to do a copy. If the buffer is not contiguous, we can use the copy to make it so.
        with MappedArray(self, name) as m:
            if m.array.data.c_contiguous:
                return np.copy(m.array)
            else:
                return np.ascontiguousarray(m.array)

    def make_image(self, name: str, width: Optional[int] = None, height: Optional[int] = None) -> Image.Image:
        """Make a PIL image from the named stream's buffer."""
        config = self.config.get(name, None)
        if config is None:
            raise RuntimeError(f'Stream {name!r} is not defined')

        fmt = config['format']
        if fmt == 'MJPEG':
            return Image.open(io.BytesIO(self.make_buffer(name)))
        mode = self.picam2.helpers._get_pil_mode(fmt)

        with MappedArray(self, name, write=False) as m:
            shape = m.array.shape
            # At this point, array is the same memory as the camera buffer - no copy has happened.
            buffer = m.array
            stride = m.array.strides[0]
            if mode == "RGBX":
                # FOr RGBX mode only, PIL shares the underlying buffer. So if we don't want to pass
                # out a handle to the camera buffer, then we must copy it.
                buffer = np.copy(m.array)
                stride = buffer.strides[0]
            elif not m.array.data.c_contiguous:
                # PIL will accept images with padding, but we must give it a contiguous buffer and
                # pass the stride as the penultimate "magic" parameter to frombuffer.
                buffer = m.array.base

            pil_img = Image.frombuffer("RGB", (shape[1], shape[0]), buffer, "raw", mode, stride, 1)

        width = width or shape[1]
        height = height or shape[0]
        if width != shape[1] or height != shape[0]:
            # This will be slow. Consider requesting camera images of this size in the first place!
            pil_img = pil_img.resize((width, height))
        return pil_img

    def save(self, name: str, file_output: Any, format: Optional[str] = None,
             exif_data: Optional[Dict[str, Any]] = None) -> None:
        """
        Save a JPEG or PNG image of the named stream's buffer.

        exif_data - dictionary containing user defined exif data (based on `piexif`). This will
            overwrite existing exif information generated by picamera2.
        """
        # We have a more optimised path for writing JPEGs using simplejpeg.
        config = self.config.get(name, None)
        if config is None:
            raise RuntimeError(f'Stream {name!r} is not defined')
        if (config['format'] == 'YUV420' or (self.FASTER_JPEG and config['format'] != "MJPEG")) and \
           self.picam2.helpers._get_format_str(file_output, format) in ('jpg', 'jpeg'):
            quality = self.picam2.options.get("quality", 90)
            with MappedArray(self, 'main') as m:
                format = self.config[name]["format"]
                if format == 'YUV420':
                    width, height = self.config[name]['size']
                    Y = m.array[:height, :width]
                    reshaped = m.array.reshape((m.array.shape[0] * 2, m.array.strides[0] // 2))
                    U = reshaped[2 * height: 2 * height + height // 2, :width // 2]
                    V = reshaped[2 * height + height // 2:, :width // 2]
                    output_bytes = simplejpeg.encode_jpeg_yuv_planes(Y, U, V, quality)
                    Y = reshaped = U = V = None
                else:
                    FORMAT_TABLE = {"XBGR8888": "RGBX", "XRGB8888": "BGRX", "BGR888": "RGB", "RGB888": "BGR"}
                    output_bytes = simplejpeg.encode_jpeg(m.array, quality, FORMAT_TABLE[format], '420')

            exif = self.picam2.helpers._prepare_exif(self.get_metadata(), exif_data)

            if isinstance(file_output, io.BytesIO):
                f = file_output
            else:
                f = open(file_output, 'wb')
            try:
                if exif:
                    # Splice in the exif data as we write it out.
                    f.write(output_bytes[:2] + bytes.fromhex('ffe1') + (len(exif) + 2).to_bytes(2, 'big'))
                    f.write(exif)
                    f.write(output_bytes[2:])
                else:
                    f.write(output_bytes)
            except Exception:
                if f is not file_output:
                    f.close()
        else:
            return self.picam2.helpers.save(self.make_image(name), self.get_metadata(), file_output,
                                            format, exif_data)

    def save_dng(self, file_output: Any, name: str = "raw") -> None:
        """Save a DNG RAW image of the raw stream's buffer."""
        # Don't use make_buffer(), this saves a copy.
        if self.stream_map.get(name) is None:
            raise RuntimeError(f'Stream {name!r} is not defined')
        with _MappedBuffer(self, name, write=False) as b:
            buffer = np.array(b, copy=False, dtype=np.uint8)
            return self.picam2.helpers.save_dng(buffer, self.get_metadata(), self.config[name], file_output)


class Helpers:
    """This class implements functionality required by the CompletedRequest methods.

    In such a way that it can be usefully accessed even without a CompletedRequest object.
    """

    def __init__(self, picam2: "Picamera2"):
        self.picam2 = picam2

    def _make_array_shared(self, buffer: np.ndarray, config: Dict[str, Any]) -> np.ndarray:
        """Makes a 2d numpy array from the named stream's buffer without copying memory.

        This method makes an array that is guaranteed to be shared with the underlying
        buffer, that is, no copy of the pixel data is made.
        """
        array = buffer
        fmt = config["format"]
        w, h = config["size"]
        stride = config["stride"]

        # Reshape the 1d array into an image, and "slice" off any padding bytes on the
        # right hand edge (which doesn't copy the pixel data).
        if fmt in ("BGR888", "RGB888"):
            if stride != w * 3:
                array = array.reshape((h, stride))
                array = array[:, :w * 3]
            image = array.reshape((h, w, 3))
        elif fmt in ("XBGR8888", "XRGB8888"):
            if stride != w * 4:
                array = array.reshape((h, stride))
                array = array[:, :w * 4]
            image = array.reshape((h, w, 4))
        elif fmt in ("BGR161616", "RGB161616"):
            if stride != w * 6:
                array = array.reshape((h, stride))
                array = array[:, :w * 6]
            array = array.view(np.uint16)
            image = array.reshape((h, w, 3))
        elif fmt in ("YUV420", "YVU420"):
            # Returning YUV420 as an image of 50% greater height (the extra bit continaing
            # the U/V data) is useful because OpenCV can convert it to RGB for us quite
            # efficiently. We leave any packing in there, however, as it would be easier
            # to remove that after conversion to RGB (if that's what the caller does).
            image = array.reshape((h * 3 // 2, stride))
        elif fmt in ("YUYV", "YVYU", "UYVY", "VYUY"):
            # These dimensions seem a bit strange, but mean that
            # cv2.cvtColor(image, cv2.COLOR_YUV2BGR_YUYV) will convert directly to RGB.
            image = array.reshape(h, stride // 2, 2)
        elif fmt == "MJPEG":
            image = np.array(Image.open(io.BytesIO(array)))  # type: ignore
        elif formats.is_raw(fmt):
            image = array.reshape((h, stride))
        else:
            raise RuntimeError("Format " + fmt + " not supported")
        return image

    def make_array(self, buffer, config):
        """Makes a 2d numpy array for the named stream's buffer.

        This method makes a copy of the underlying camera buffer, so that it can be
        safely returned to the camera system.
        """
        array = self._make_array_shared(buffer, config)
        if array.data.c_contiguous:
            return np.copy(array)
        else:
            return np.ascontiguousarray(array)

    def _get_pil_mode(self, fmt):
        mode_lookup = {"RGB888": "BGR", "BGR888": "RGB", "XBGR8888": "RGBX", "XRGB8888": "BGRX"}
        mode = mode_lookup.get(fmt, None)
        if mode is None:
            raise RuntimeError(f"Stream format {fmt} not supported for PIL images")
        return mode

    def make_image(self, buffer: np.ndarray, config: Dict[str, Any], width: Optional[int] = None,
                   height: Optional[int] = None) -> Image.Image:
        """Make a PIL image from the named stream's buffer."""
        fmt = config["format"]
        if fmt == "MJPEG":
            return Image.open(io.BytesIO(buffer))  # type: ignore
        else:
            rgb = self._make_array_shared(buffer, config)

        # buffer was already a copy, so don't need to worry about an extra copy for the "RGBX" mode.
        buf = rgb
        if not rgb.data.c_contiguous:
            buf = rgb.base

        mode = self._get_pil_mode(fmt)
        pil_img = Image.frombuffer("RGB", (rgb.shape[1], rgb.shape[0]), buf, "raw", mode, rgb.strides[0], 1)

        width = width or rgb.shape[1]
        height = height or rgb.shape[0]
        if width != rgb.shape[1] or height != rgb.shape[0]:
            # This will be slow. Consider requesting camera images of this size in the first place!
            pil_img = pil_img.resize((width, height))  # type: ignore
        return pil_img

    def _prepare_exif(self, metadata, exif_data):
        exif = b''
        if "AnalogueGain" in metadata and "DigitalGain" in metadata:
            datetime_now = datetime.now().strftime("%Y:%m:%d %H:%M:%S")
            zero_ifd = {piexif.ImageIFD.Make: "Raspberry Pi",
                        piexif.ImageIFD.Model: self.picam2.camera.id,
                        piexif.ImageIFD.Software: "Picamera2",
                        piexif.ImageIFD.DateTime: datetime_now}
            total_gain = metadata["AnalogueGain"] * metadata["DigitalGain"]
            exif_ifd = {piexif.ExifIFD.DateTimeOriginal: datetime_now,
                        piexif.ExifIFD.ExposureTime: (metadata["ExposureTime"], 1000000),
                        piexif.ExifIFD.ISOSpeedRatings: int(total_gain * 100)}
            exif_dict = {"0th": zero_ifd, "Exif": exif_ifd}
            # merge user provided exif data, overwriting the defaults
            exif_dict = exif_dict | (exif_data or {})
            exif = piexif.dump(exif_dict)
        return exif

    def _get_format_str(self, file_output, format):
        if isinstance(format, str):
            return format.lower()
        elif isinstance(file_output, str):
            return file_output.split('.')[-1].lower()
        elif isinstance(file_output, Path):
            return file_output.suffix.lower()
        else:
            raise RuntimeError("Cannot determine format to save")

    def save(self, img: Image.Image, metadata: Dict[str, Any], file_output: Union[str, Path], format: Optional[str] = None,
             exif_data: Optional[Dict] = None) -> None:
        """Save a JPEG or PNG image of the named stream's buffer.

        exif_data - dictionary containing user defined exif data (based on `piexif`). This will
            overwrite existing exif information generated by picamera2.
        """
        if exif_data is None:
            exif_data = {}
        # This is probably a hideously expensive way to do a capture.
        start_time = time.monotonic()
        format_str = self._get_format_str(file_output, format)
        if format_str in ('png') and img.mode == 'RGBX':
            # It seems we can't save an RGBX png file, so make it RGBA instead. We can't use RGBA
            # everywhere, because we can only save an RGBX jpeg, not an RGBA one.
            img = img.convert(mode='RGBA')
        exif = b''
        if format_str in ('jpg', 'jpeg'):
            # Make up some extra EXIF data.
            exif = self._prepare_exif(metadata, exif_data)
        # compress_level=1 saves pngs much faster, and still gets most of the compression.
        png_compress_level = self.picam2.options.get("compress_level", 1)
        jpeg_quality = self.picam2.options.get("quality", 90)
        keywords = {"compress_level": png_compress_level, "quality": jpeg_quality, "format": format}
        if exif != b'':
            keywords |= {"exif": exif}
        img.save(file_output, **keywords)
        end_time = time.monotonic()
        _log.info(f"Saved {self} to file {file_output}.")
        _log.info(f"Time taken for encode: {(end_time-start_time)*1000} ms.")

    def save_dng(self, buffer: np.ndarray, metadata: Dict[str, Any], config: Dict[str, Any], file_output: Any) -> None:
        """Save a DNG RAW image of the raw stream's buffer."""
        start_time = time.monotonic()
        raw = self.make_array(buffer, config)
        config = config.copy()

        fmt = SensorFormat(config['format'])
        if fmt.packing == 'PISP_COMP1':
            raw = self.decompress(raw)
            fmt.bit_depth = 16
            config['format'] = fmt.unpacked
            config['stride'] = raw.shape[1]
            config['framesize'] = raw.shape[0] * raw.shape[1]

        camera = Picamera2Camera(config, metadata)
        r = PICAM2DNG(camera)

        dng_compress_level = self.picam2.options.get("compress_level", 0)

        r.options(compress=dng_compress_level)
        # PiDNG doesn't accpet a BytesIO, but returns a byte array if the filename is empty.
        if isinstance(file_output, io.BytesIO):
            buf = r.convert(raw, "")
            file_output.write(buf)
        else:
            r.convert(raw, str(file_output))

        end_time = time.monotonic()
        _log.info(f"Saved {self} to file {file_output}.")
        _log.info(f"Time taken for encode: {(end_time-start_time)*1000} ms.")

    def decompress(self, array: np.ndarray):
        """Decompress an image buffer that has been compressed with a PiSP compression format."""
        # These are the standard configurations used in the drivers.
        offset = 2048

        words = array.view(np.int32)  # Assume all Pis are little-endian. Note signed arithmetic is used!
        words = words.reshape((words.shape[0], words.shape[1] // 2, 2))  # pairs of words by component
        qmode = words & 3
        pix0 = (words >> 2) & 511
        pix1 = ((words >> 11) & 127) - 64
        pix2 = (words >> 18) & 127
        pix3 = (words >> 25) & 127
        q1 = np.copy(pix0)
        q2 = pix1 + 448
        np.maximum(pix0, pix0 - pix1, where=(qmode * pix0 < 768), out=q1)
        np.maximum(pix0, pix0 + pix1, where=(qmode * pix0 < 768), out=q2)
        q0 = np.minimum(1536 >> qmode, np.maximum(0, q1 - 64)) + pix2
        q3 = np.minimum(1536 >> qmode, np.maximum(0, q2 - 64)) + pix3
        np.maximum(np.maximum(16 * q0, 32 * (q0 - 160)), 64 * qmode * q0, out=pix0)
        np.maximum(np.maximum(16 * q1, 32 * (q1 - 160)), 64 * qmode * q1, out=pix1)
        np.maximum(np.maximum(16 * q2, 32 * (q2 - 160)), 64 * qmode * q2, out=pix2)
        np.maximum(np.maximum(16 * q3, 32 * (q3 - 160)), 64 * qmode * q3, out=pix3)
        q2 = (words >> 2) & 32767
        q3 = (words >> 17) & 32767
        q0 = (q2 & 15) + 16 * ((q2 >> 8) // 11)
        q1 = (q2 >> 4) % 176
        q2 = (q3 & 15) + 16 * ((q3 >> 8) // 11)
        q3 = (q3 >> 4) % 176
        np.maximum(256 * q0, 512 * (q0 - 47), out=pix0, where=(qmode == 3))
        np.maximum(256 * q1, 512 * (q1 - 47), out=pix1, where=(qmode == 3))
        np.maximum(256 * q2, 512 * (q2 - 47), out=pix2, where=(qmode == 3))
        np.maximum(256 * q3, 512 * (q3 - 47), out=pix3, where=(qmode == 3))
        res = np.stack((pix0, pix1, pix2, pix3), axis=2).reshape(array.shape)

        res = np.clip(res + offset, 0, 65535).astype(np.uint16)
        return res.view(np.uint8)



import logging
from functools import lru_cache
from operator import attrgetter

import numpy as np
from libcamera import Transform

try:
    import cv2
    cv2_available = True
except ImportError:
    cv2_available = False

from .qt_compatibility import _QT_BINDING, _get_qt_modules


@lru_cache(maxsize=None, typed=False)
def _get_qpicamera2(qt_module: _QT_BINDING):

    # Get Qt modules
    QtCore, QtGui, QtWidgets = _get_qt_modules(qt_module)

    # Get from QtCore
    QRect, QRectF, QSize, QSocketNotifier, Qt, pyqtSignal, pyqtSlot = attrgetter(
        'QRect', 'QRectF', 'QSize', 'QSocketNotifier', 'Qt', 'pyqtSignal', 'pyqtSlot')(QtCore)

    # Get from QtGui
    QBrush, QColor, QImage, QPixmap, QTransform = attrgetter(
        'QBrush', 'QColor', 'QImage', 'QPixmap', 'QTransform')(QtGui)

    # Get from QtWidgets
    QGraphicsScene, QGraphicsView = attrgetter(
        'QGraphicsScene', 'QGraphicsView')(QtWidgets)

    class QPicamera2(QGraphicsView):
        done_signal = pyqtSignal(object)
        update_overlay_signal = pyqtSignal(object)

        def __init__(self, picam2, parent=None, width=640, height=480, bg_colour=(20, 20, 20),
                     keep_ar=True, transform=None, preview_window=None):
            super().__init__(parent=parent)
            self.picamera2 = picam2
            picam2.attach_preview(preview_window)
            self.preview_window = preview_window
            self.keep_ar = keep_ar
            self.transform = Transform() if transform is None else transform
            self.image_size = None
            self.last_rect = QRect(0, 0, 0, 0)

            self.size = QSize(width, height)
            self.pixmap = None
            self.overlay = None
            self.scene = QGraphicsScene()
            self.setScene(self.scene)
            self.setBackgroundBrush(QBrush(QColor(*bg_colour)))
            self.resize(width, height)
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            self.enabled = True
            self.title_function = None

            self.update_overlay_signal.connect(self.update_overlay)
            self.camera_notifier = QSocketNotifier(self.picamera2.notifyme_r,
                                                   QSocketNotifier.Type.Read, self)
            self.camera_notifier.activated.connect(self.handle_requests)
            # Must always run cleanup when this widget goes away.
            self.destroyed.connect(lambda: self.cleanup())
            self.running = True

        def cleanup(self):
            if not self.running:
                return
            self.running = False
            del self.scene
            del self.overlay
            self.camera_notifier.deleteLater()
            # We have to tell both the preview window and the Picamera2 object that we have
            # disappeared.
            self.picamera2.detach_preview()
            if self.preview_window is not None:  # will be none when a proper Qt app
                self.preview_window.qpicamera2 = None

        def closeEvent(self, event):
            self.cleanup()

        def signal_done(self, job):
            self.done_signal.emit(job)

        def image_dimensions(self):
            # The dimensions of the camera images we're displaying.
            camera_config = self.picamera2.camera_config
            if camera_config and camera_config['display'] is not None:
                # This works even before we receive any camera images.
                size = (self.picamera2.stream_map[camera_config['display']].configuration.size.width,
                        self.picamera2.stream_map[camera_config['display']].configuration.size.height)
            elif self.image_size is not None:
                # If the camera is unconfigured, stick with the last size (if available).
                size = self.image_size
            else:
                # Otherwise pretend the image fills everything.
                rect = self.viewport().rect()
                size = rect.width(), rect.height()
            self.image_size = size
            return size

        def set_overlay(self, overlay):
            camera_config = self.picamera2.camera_config
            if not camera_config:
                raise RuntimeError("Camera must be configured before using set_overlay")

            new_pixmap = None
            if overlay is not None:
                overlay = np.copy(overlay, order='C')
                shape = overlay.shape
                qim = QImage(overlay.data, shape[1], shape[0], QImage.Format.Format_RGBA8888)
                new_pixmap = QPixmap(qim)
                # No scaling here - we leave it to fitInView to set that up.
            self.update_overlay_signal.emit(new_pixmap)

        @pyqtSlot(object)
        def update_overlay(self, pix):
            if pix is None:
                # Delete overlay if present
                if self.overlay is not None:
                    self.scene.removeItem(self.overlay)
                    self.overlay = None
                    return
            elif self.overlay is None:
                # Need to add the overlay to the scene
                self.overlay = self.scene.addPixmap(pix)
                self.overlay.setZValue(100)
            else:
                # Just update it
                self.overlay.setPixmap(pix)
            self.fitInView()

        @pyqtSlot(bool)
        def set_enabled(self, enabled):
            self.enabled = enabled

        def fitInView(self):
            # Reimplemented fitInView to remove fixed border
            image_w, image_h = self.image_dimensions()
            rect = QRectF(0, 0, image_w, image_h)
            self.setSceneRect(rect)
            # I get one column of background peeping through on the right without this:
            viewrect = self.viewport().rect().adjusted(0, 0, 1, 1)
            self.resetTransform()
            factor_x = viewrect.width() / image_w
            factor_y = viewrect.height() / image_h
            if self.keep_ar:
                factor_x = min(factor_x, factor_y)
                factor_y = factor_x
            if self.transform.hflip:
                factor_x = -factor_x
            if self.transform.vflip:
                factor_y = -factor_y
            self.scale(factor_x, factor_y)

            # This scales the overlay to be on top of the camera image.
            if self.overlay:
                rect = self.overlay.boundingRect()
                self.overlay.resetTransform()
                factor_x = image_w / rect.width()
                factor_y = image_h / rect.height()
                translate_x, translate_y = 0, 0
                if self.transform.hflip:
                    factor_x = -factor_x
                    translate_x = -rect.width()
                if self.transform.vflip:
                    factor_y = -factor_y
                    translate_y = -rect.height()
                transform = QTransform.fromScale(factor_x, factor_y)
                transform.translate(translate_x, translate_y)
                self.overlay.setTransform(transform, True)

        def resizeEvent(self, event):
            self.fitInView()

        def render_request(self, completed_request):
            """Draw the camera image using Qt."""
            if not self.enabled:
                return

            if self.title_function is not None:
                self.setWindowTitle(self.title_function(completed_request.get_metadata()))

            camera_config = completed_request.config
            display_stream_name = camera_config['display']
            stream_config = camera_config[display_stream_name]

            img = completed_request.make_array(display_stream_name)
            if stream_config["format"] in ("YUV420", "YUYV"):
                if cv2_available:
                    if stream_config["format"] == "YUV420":
                        img = cv2.cvtColor(img, cv2.COLOR_YUV420p2BGR)
                    else:
                        img = cv2.cvtColor(img, cv2.COLOR_YUV2RGB_YUYV)
                else:
                    logging.error("Qt preview cannot display YUV420/YUYV without cv2")
                    return

            # Crop width for two reasons: (1) to remove "stride" padding from YUV images;
            # (2) to ensure the RGB buffer passed to QImage has mandatory 4-byte alignment.
            # [TODO: Consider QImage.Format_RGB32, if byte order can be made correct]
            width = min(img.shape[1], stream_config["size"][0])
            width -= width % 4
            img = np.ascontiguousarray(img[:, :width, :3])
            if stream_config['format'] in ('RGB888', 'XRGB8888'):
                fmt = QImage.Format.Format_BGR888
            else:
                fmt = QImage.Format.Format_RGB888
            qim = QImage(img.data, width, img.shape[0], fmt)
            pix = QPixmap(qim)
            # Add the pixmap to the scene if there wasn't one, or replace it if the images have
            # changed size.
            if self.pixmap is None or pix.rect() != self.last_rect:
                if self.pixmap:
                    self.scene.removeItem(self.pixmap)
                self.last_rect = pix.rect()
                self.pixmap = self.scene.addPixmap(pix)
                self.fitInView()
            else:
                # Update pixmap
                self.pixmap.setPixmap(pix)

        @pyqtSlot()
        def handle_requests(self):
            if not self.running:
                return
            self.picamera2.notifymeread.read()
            self.picamera2.process_requests(self)

    return QPicamera2


# Define QPicamera2 class for compatability
try:
    QPicamera2 = _get_qpicamera2(_QT_BINDING.PyQt5)
except ImportError:
    pass



import os
import threading
import time
from functools import lru_cache
from operator import attrgetter

os.environ["PYOPENGL_PLATFORM"] = "egl"

from libcamera import Transform
from OpenGL.EGL.EXT.image_dma_buf_import import *
from OpenGL.EGL.KHR.image import *
from OpenGL.EGL.VERSION.EGL_1_0 import *
from OpenGL.EGL.VERSION.EGL_1_2 import *
from OpenGL.EGL.VERSION.EGL_1_3 import *
from OpenGL.GL import shaders
from OpenGL.GLES2.OES.EGL_image import *
from OpenGL.GLES2.OES.EGL_image_external import *
from OpenGL.GLES2.VERSION.GLES2_2_0 import *
from OpenGL.GLES3.VERSION.GLES3_3_0 import *

from picamera2.previews.gl_helpers import *

from .qt_compatibility import _QT_BINDING, _get_qt_modules


class EglState:
    def __init__(self):
        self.create_display()
        self.choose_config()
        self.create_context()
        # GLES version earlier than 3.0 (e.g. on a Pi 3) don't support this, but it's
        # only a check so we can skip it.
        # check_gl_extensions(["GL_OES_EGL_image"])
        n = GLint()
        glGetIntegerv(GL_MAX_TEXTURE_SIZE, n)
        self.max_texture_size = n.value

    def create_display(self):
        xdisplay = getEGLNativeDisplay()
        self.display = eglGetDisplay(xdisplay)

    def choose_config(self):
        major, minor = EGLint(), EGLint()

        eglInitialize(self.display, major, minor)

        check_egl_extensions(self.display, ["EGL_EXT_image_dma_buf_import"])

        eglBindAPI(EGL_OPENGL_ES_API)

        config_attribs = [
            EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
            EGL_RED_SIZE, 8,
            EGL_GREEN_SIZE, 8,
            EGL_BLUE_SIZE, 8,
            EGL_ALPHA_SIZE, 0,
            EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
            EGL_NONE,
        ]

        n = EGLint()
        configs = (EGLConfig * 1)()
        eglChooseConfig(self.display, config_attribs, configs, 1, n)
        self.config = configs[0]

    def create_context(self):
        context_attribs = [
            EGL_CONTEXT_CLIENT_VERSION, 2,
            EGL_NONE,
        ]

        self.context = eglCreateContext(self.display, self.config, EGL_NO_CONTEXT, context_attribs)
        eglMakeCurrent(self.display, EGL_NO_SURFACE, EGL_NO_SURFACE, self.context)


@lru_cache(maxsize=None, typed=False)
def _get_qglpicamera2(qt_module: _QT_BINDING):

    # Get Qt modules
    QtCore, QtGui, QtWidgets = _get_qt_modules(qt_module)

    # Get from QtCore
    QSocketNotifier, Qt, pyqtSignal, pyqtSlot = attrgetter(
        'QSocketNotifier', 'Qt', 'pyqtSignal', 'pyqtSlot')(QtCore)

    # Get from QtWidgets
    QWidget = attrgetter(
        'QWidget')(QtWidgets)

    class QGlPicamera2(QWidget):
        done_signal = pyqtSignal(object)

        def __init__(self, picam2, parent=None, width=640, height=480, bg_colour=(20, 20, 20),
                     keep_ar=True, transform=None, preview_window=None):
            super().__init__(parent=parent)
            self.resize(width, height)

            self.setAttribute(Qt.WidgetAttribute.WA_PaintOnScreen)
            self.setAttribute(Qt.WidgetAttribute.WA_NativeWindow)

            self.bg_colour = [colour / 255.0 for colour in bg_colour] + [1.0]
            self.keep_ar = keep_ar
            self.transform = Transform() if transform is None else transform
            self.lock = threading.Lock()
            self.count = 0
            self.overlay_present = False
            self.buffers = {}
            self.surface = None
            self.current_request = None
            self.own_current = False
            self.stop_count = 0
            self.title_function = None
            self.egl = EglState()
            if picam2.verbose_console:
                print(f"EGL {eglQueryString(self.egl.display, EGL_VENDOR).decode()} "
                      f"{eglQueryString(self.egl.display, EGL_VERSION).decode()}")
            self.init_gl()

            # set_overlay could be called before the first frame arrives, hence:
            eglMakeCurrent(self.egl.display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT)

            self.picamera2 = picam2
            picam2.attach_preview(preview_window)
            self.preview_window = preview_window

            self.camera_notifier = QSocketNotifier(self.picamera2.notifyme_r,
                                                   QSocketNotifier.Type.Read, self)
            self.camera_notifier.activated.connect(self.handle_requests)
            # Must always run cleanup when this widget goes away.
            self.destroyed.connect(lambda: self.cleanup())
            self.running = True

        def cleanup(self):
            if not self.running:
                return
            self.running = False
            self.camera_notifier.deleteLater()
            eglDestroySurface(self.egl.display, self.surface)
            self.surface = None
            # We may be hanging on to a request, return it to the camera system.
            if self.current_request is not None and self.own_current:
                self.current_request.release()
            self.current_request = None
            # We have to tell both the preview window and the Picamera2 object that we have
            # disappeared.
            self.picamera2.detach_preview()
            if self.preview_window is not None:  # will be none when a proper Qt app
                self.preview_window.qpicamera2 = None

            # Some extra EGL cleanup seems to be required.
            for (_, buffer) in self.buffers.items():
                glDeleteTextures(1, [buffer.texture])
            self.buffers = {}
            eglDestroyContext(self.egl.display, self.egl.context)
            self.egl.context = None

        def closeEvent(self, event):
            self.cleanup()

        def signal_done(self, job):
            self.done_signal.emit(job)

        def paintEngine(self):
            return None

        def create_surface(self):
            native_surface = c_void_p(self.winId().__int__())
            surface = eglCreateWindowSurface(self.egl.display, self.egl.config,
                                             native_surface, None)

            self.surface = surface

        def init_gl(self):
            self.create_surface()

            eglMakeCurrent(self.egl.display, self.surface, self.surface, self.egl.context)

            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
            self.overlay_texture = glGenTextures(1)

            vertShaderSrc_image = f"""
                attribute vec2 aPosition;
                varying vec2 texcoord;

                void main()
                {{
                    gl_Position = vec4(aPosition * 2.0 - 1.0, 0.0, 1.0);
                    texcoord.x = {'1.0 - ' if self.transform.hflip else ''}aPosition.x;
                    texcoord.y = {'' if self.transform.vflip else '1.0 - '}aPosition.y;
                }}
            """
            fragShaderSrc_image = """
                #extension GL_OES_EGL_image_external : enable
                precision mediump float;
                varying vec2 texcoord;
                uniform samplerExternalOES texture;

                void main()
                {
                    gl_FragColor = texture2D(texture, texcoord);
                }
            """
            vertShaderSrc_overlay = """
                attribute vec2 aPosition;
                varying vec2 texcoord;

                void main()
                {
                    gl_Position = vec4(aPosition * 2.0 - 1.0, 0.0, 1.0);
                    texcoord.x = aPosition.x;
                    texcoord.y = 1.0 - aPosition.y;
                }
            """
            fragShaderSrc_overlay = """
                precision mediump float;
                varying vec2 texcoord;
                uniform sampler2D overlay;

                void main()
                {
                    gl_FragColor = texture2D(overlay, texcoord);
                }
            """

            vertex_shader = shaders.compileShader(vertShaderSrc_image, GL_VERTEX_SHADER),
            # For some reason I seem to be getting a 1 element tuple back. Absolutely no clue why.
            if isinstance(vertex_shader, tuple):
                vertex_shader = vertex_shader[0]
            fragment_shader = shaders.compileShader(fragShaderSrc_image, GL_FRAGMENT_SHADER)
            self.program_image = shaders.compileProgram(vertex_shader, fragment_shader)

            self.program_overlay = shaders.compileProgram(
                shaders.compileShader(vertShaderSrc_overlay, GL_VERTEX_SHADER),
                shaders.compileShader(fragShaderSrc_overlay, GL_FRAGMENT_SHADER)
            )

            vertPositions = [
                0.0, 0.0,
                1.0, 0.0,
                1.0, 1.0,
                0.0, 1.0
            ]

            inputAttrib = glGetAttribLocation(self.program_image, "aPosition")
            glVertexAttribPointer(inputAttrib, 2, GL_FLOAT, GL_FALSE, 0, vertPositions)
            glEnableVertexAttribArray(inputAttrib)

            inputAttrib = glGetAttribLocation(self.program_overlay, "aPosition")
            glVertexAttribPointer(inputAttrib, 2, GL_FLOAT, GL_FALSE, 0, vertPositions)
            glEnableVertexAttribArray(inputAttrib)

            glUseProgram(self.program_overlay)
            glUniform1i(glGetUniformLocation(self.program_overlay, "overlay"), 0)

        class Buffer:
            # libcamera format string -> DRM fourcc, note that 24-bit formats are not supported
            FMT_MAP = {
                "XRGB8888": "XR24",
                "XBGR8888": "XB24",
                "YUYV": "YUYV",
                # doesn't work "YVYU": "YVYU",
                "UYVY": "UYVY",
                # doesn't work "VYUY": "VYUY",
                "YUV420": "YU12",
                "YVU420": "YV12",
            }

            def __init__(self, display, completed_request, max_texture_size):
                picam2 = completed_request.picam2
                stream = picam2.stream_map[picam2.display_stream_name]
                fb = completed_request.request.buffers[stream]

                cfg = stream.configuration
                pixel_format = str(cfg.pixel_format)
                if pixel_format not in self.FMT_MAP:
                    raise RuntimeError(f"Format {pixel_format} not supported by QGlPicamera2 preview")
                fmt = str_to_fourcc(self.FMT_MAP[pixel_format])
                w, h = (cfg.size.width, cfg.size.height)
                if w > max_texture_size or h > max_texture_size:
                    raise RuntimeError(f"Maximum supported preview image size is {max_texture_size}")
                if pixel_format in ("YUV420", "YVU420"):
                    h2 = h // 2
                    stride2 = cfg.stride // 2
                    attribs = [
                        EGL_WIDTH, w,
                        EGL_HEIGHT, h,
                        EGL_LINUX_DRM_FOURCC_EXT, fmt,
                        EGL_DMA_BUF_PLANE0_FD_EXT, fb.planes[0].fd,
                        EGL_DMA_BUF_PLANE0_OFFSET_EXT, 0,
                        EGL_DMA_BUF_PLANE0_PITCH_EXT, cfg.stride,
                        EGL_DMA_BUF_PLANE1_FD_EXT, fb.planes[0].fd,
                        EGL_DMA_BUF_PLANE1_OFFSET_EXT, h * cfg.stride,
                        EGL_DMA_BUF_PLANE1_PITCH_EXT, stride2,
                        EGL_DMA_BUF_PLANE2_FD_EXT, fb.planes[0].fd,
                        EGL_DMA_BUF_PLANE2_OFFSET_EXT, h * cfg.stride + h2 * stride2,
                        EGL_DMA_BUF_PLANE2_PITCH_EXT, stride2,
                        EGL_NONE,
                    ]
                else:
                    attribs = [
                        EGL_WIDTH, w,
                        EGL_HEIGHT, h,
                        EGL_LINUX_DRM_FOURCC_EXT, fmt,
                        EGL_DMA_BUF_PLANE0_FD_EXT, fb.planes[0].fd,
                        EGL_DMA_BUF_PLANE0_OFFSET_EXT, 0,
                        EGL_DMA_BUF_PLANE0_PITCH_EXT, cfg.stride,
                        EGL_NONE,
                    ]

                image = eglCreateImageKHR(display,
                                          EGL_NO_CONTEXT,
                                          EGL_LINUX_DMA_BUF_EXT,
                                          None,
                                          attribs)

                self.texture = glGenTextures(1)
                glBindTexture(GL_TEXTURE_EXTERNAL_OES, self.texture)
                glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
                glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
                glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
                glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
                glEGLImageTargetTexture2DOES(GL_TEXTURE_EXTERNAL_OES, image)

                eglDestroyImageKHR(display, image)

        def set_overlay(self, overlay):
            if not self.picamera2.camera_config:
                raise RuntimeError("Camera must be configured before setting overlay")
            if self.picamera2.camera_config['buffer_count'] < 2:
                raise RuntimeError("Need at least buffer_count=2 to set overlay")

            with self.lock:
                eglMakeCurrent(self.egl.display, self.surface, self.surface, self.egl.context)

                if overlay is None:
                    self.overlay_present = False
                    self.repaint(self.current_request)
                else:
                    # All this swapping round of contexts is a bit icky, but I'd rather copy
                    # the overlay here so that the user doesn't have to worry about us still
                    # using it after this function returns.
                    glBindTexture(GL_TEXTURE_2D, self.overlay_texture)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
                    (height, width, channels) = overlay.shape
                    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, overlay)
                    self.overlay_present = True
                    self.repaint(self.current_request)

                eglMakeCurrent(self.egl.display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT)

        def repaint(self, completed_request, update_viewport=False):
            # The context should be set up and cleared by the caller.
            if completed_request and completed_request.request not in self.buffers:
                if self.stop_count != self.picamera2.stop_count:
                    if self.picamera2.verbose_console:
                        print("Garbage collect", len(self.buffers), "textures")
                    for (_, buffer) in self.buffers.items():
                        glDeleteTextures(1, [buffer.texture])
                    self.buffers = {}
                    self.stop_count = self.picamera2.stop_count

                if self.picamera2.verbose_console:
                    print("Make buffer for request", completed_request.request)
                self.buffers[completed_request.request] = self.Buffer(
                    self.egl.display, completed_request, self.egl.max_texture_size)

                # New buffers mean the image size may change so update the viewport just in case.
                update_viewport = True

            # If there's no request, then the viewport may never have been set up, so force it anyway.
            if update_viewport or not completed_request:
                x_off, y_off, w, h = self.recalculate_viewport()
                glViewport(x_off, y_off, w, h)

            glClearColor(*self.bg_colour)
            glClear(GL_COLOR_BUFFER_BIT)

            if completed_request:
                buffer = self.buffers[completed_request.request]
                glUseProgram(self.program_image)
                glBindTexture(GL_TEXTURE_EXTERNAL_OES, buffer.texture)
                glDrawArrays(GL_TRIANGLE_FAN, 0, 4)

            if self.overlay_present:
                glUseProgram(self.program_overlay)
                glBindTexture(GL_TEXTURE_2D, self.overlay_texture)
                glDrawArrays(GL_TRIANGLE_FAN, 0, 4)

            eglSwapBuffers(self.egl.display, self.surface)

        def render_request(self, completed_request):
            """Draw the camera image using Qt and OpenGL/GLES."""
            # For reasons not terribly well understood, eglMakeCurrent hangs in the X-Wayland world if the
            # window is being closed. This appears to stop it.
            if not self.isVisible():
                return
            if self.title_function is not None:
                self.setWindowTitle(self.title_function(completed_request.get_metadata()))
            with self.lock:
                if self.running:
                    eglMakeCurrent(self.egl.display, self.surface, self.surface, self.egl.context)
                    self.repaint(completed_request)
                    eglMakeCurrent(self.egl.display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT)
                if self.current_request and self.own_current:
                    self.current_request.release()
                self.current_request = completed_request
                self.own_current = (completed_request.config['buffer_count'] > 1)
                if self.own_current:
                    self.current_request.acquire()

        @pyqtSlot()
        def handle_requests(self):
            if not self.running:
                return
            self.picamera2.notifymeread.read()
            self.picamera2.process_requests(self)

        def recalculate_viewport(self):
            window_w = self.width()
            window_h = self.height()

            stream_map = self.picamera2.stream_map
            camera_config = self.picamera2.camera_config
            if not self.keep_ar or not camera_config or camera_config['display'] is None:
                return 0, 0, window_w, window_h

            image_w, image_h = (stream_map[camera_config['display']].configuration.size.width,
                                stream_map[camera_config['display']].configuration.size.height)
            if image_w * window_h > window_w * image_h:
                w = window_w
                h = w * image_h // image_w
            else:
                h = window_h
                w = h * image_w // image_h
            x_off = (window_w - w) // 2
            y_off = (window_h - h) // 2
            return x_off, y_off, w, h

        def resizeEvent(self, event):
            with self.lock:
                eglMakeCurrent(self.egl.display, self.surface, self.surface, self.egl.context)
                self.repaint(self.current_request, update_viewport=True)
                eglMakeCurrent(self.egl.display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT)

        def show(self):
            super().show()
            # We seem to need a short delay before rendering the background colour works. No idea why.
            time.sleep(0.05)
            self.resizeEvent(None)

    return QGlPicamera2


# Define QGlPicamera2 class for compatability
try:
    QPicamera2 = _get_qglpicamera2(_QT_BINDING.PyQt5)
except ImportError:
    pass



import ctypes
import fcntl
import io
import json
import multiprocessing
import os
import struct
import sys
import time
from typing import List, Optional

import jsonschema
import numpy as np
from libarchive.read import fd_reader
from libcamera import Rectangle, Size
from tqdm import tqdm
from v4l2 import (VIDIOC_S_CTRL, VIDIOC_S_EXT_CTRLS, v4l2_control,
                  v4l2_ext_control, v4l2_ext_controls)

from picamera2 import CompletedRequest, Picamera2

NETWORK_NAME_LEN = 64
MAX_NUM_TENSORS = 16
MAX_NUM_DIMENSIONS = 16

FW_LOADER_STAGE = 0
FW_MAIN_STAGE = 1
FW_NETWORK_STAGE = 2

NETWORK_FW_FD_CTRL_ID = 0x00982901
ROI_CTRL_ID = 0x00982900


# struct OutputTensorInfo from libcamera
class OutputTensorInfo(ctypes.LittleEndianStructure):
    _fields_ = [
        ('tensor_data_num', ctypes.c_uint32),
        ('num_dimensions', ctypes.c_uint32),
        ('size', ctypes.c_uint16 * MAX_NUM_DIMENSIONS),
    ]


# struct CnnOutputTensorInfoExported from libcamera
class CnnOutputTensorInfoExported(ctypes.LittleEndianStructure):
    _fields_ = [
        ('network_name', ctypes.c_char * NETWORK_NAME_LEN),
        ('num_tensors', ctypes.c_uint32),
        ('info', OutputTensorInfo * MAX_NUM_TENSORS)
    ]


class NetworkIntrinsics:
    def __init__(self, val=None):
        self.__intrinsics: Optional[dict] = None
        self.__schema = {
            "$schema": "https://json-schema.org/draft-07/schema",
            "title": "network_intrinsics",
            "type": "object",
            "properties": {
                "task": {
                    "type": "string",
                    "enum": ["classification", "object detection", "pose estimation", "segmentation"],
                    "description": "Network task",
                },
                "inference_rate": {"type": "number", "minimum": 0},
                "cpu": {
                    "type": "object",
                    "properties": {
                        "bbox_normalization": {"type": "boolean"},
                        "bbox_order": {"type": "string", "enum": ["xy", "yx"]},
                        "softmax": {"type": "boolean"},
                        "post_processing": {"type": "string"},
                    },
                },
                "input_aspect_ratio": {
                    "type": "object",
                    "properties": {
                        "width": {"type": "integer", "exclusiveMinimum": 0},
                        "height": {"type": "integer", "exclusiveMinimum": 0},
                    },
                    "required": ["width", "height"],
                },
                "classes": {
                    "type": "object",
                    "properties": {
                        "labels": {"type": "array", "items": {"type": "string"}},
                        "ignore_undefined": {"type": "boolean"},
                    },
                },
            },
        }
        if val is not None:
            jsonschema.validate(val, self.__schema)
            self.__intrinsics = val

        self.__defaults = {'inference_rate': 30.0}
        jsonschema.validate(self.__defaults, self.__schema | {'additionalProperties': False})

    @property
    def intrinsics(self) -> Optional[dict]:
        return self.__intrinsics

    @intrinsics.setter
    def intrinsics(self, val):
        jsonschema.validate(val, self.__schema)
        self.__intrinsics = val

    def __repr__(self):
        return json.dumps(self.__intrinsics) if self.__intrinsics else ""

    def __top_level_validated_insert(self, val: dict):
        jsonschema.validate(val, self.__schema | {'additionalProperties': False})
        self.__intrinsics = self.__intrinsics | val if self.__intrinsics else val

    def __intrinsics_has_key(self, key: str) -> bool:
        return key in self.__intrinsics if self.__intrinsics else False

    def __intrinsics_get_key(self, key, default=None):
        return self.__intrinsics.get(key, default) if self.__intrinsics else default

    def update_with_defaults(self):
        # Updates intrinsics with default settings (but does not overwrite)
        if not self.__intrinsics:
            self.__intrinsics = {}
        self.__intrinsics = self.__defaults | self.__intrinsics

    @property
    def task(self) -> Optional[str]:
        return self.__intrinsics_get_key('task')

    @task.setter
    def task(self, val: str):
        self.__top_level_validated_insert({'task': val})

    @property
    def inference_rate(self) -> Optional[float]:
        return self.__intrinsics_get_key('inference_rate')

    @inference_rate.setter
    def inference_rate(self, val: float):
        if val < 0:
            if self.__intrinsics is not None:
                self.__intrinsics.pop('inference_rate', None)
        else:
            self.__top_level_validated_insert({'inference_rate': val})

    @property
    def fps(self) -> Optional[float]:
        # @deprecated("Prefer inference_rate")
        return self.inference_rate

    @fps.setter
    def fps(self, val: Optional[float]):
        # @deprecated("Prefer inference_rate")
        self.inference_rate = val

    def __get_cpu(self, key: str):
        return self.__intrinsics['cpu'].get(key, None) if self.__intrinsics_has_key('cpu') else None

    def __set_cpu(self, val: dict):
        jsonschema.validate({'cpu': val}, self.__schema | {'additionalProperties': False})
        cpu = self.__intrinsics_get_key('cpu', {}) | val
        if self.__intrinsics:
            self.__intrinsics['cpu'] = cpu
        else:
            self.__intrinsics = {'cpu': cpu}

    @property
    def bbox_normalization(self) -> Optional[bool]:
        return self.__get_cpu('bbox_normalization')

    @bbox_normalization.setter
    def bbox_normalization(self, val: Optional[bool]):
        if val is None:
            return

        if val:
            self.__set_cpu({'bbox_normalization': val})
        elif self.__intrinsics_has_key('cpu'):
            self.__intrinsics['cpu'].pop('bbox_normalization', None)

        if self.__intrinsics_has_key('cpu') and len(self.__intrinsics['cpu']) == 0:
            self.__intrinsics.pop('cpu')

    @property
    def bbox_order(self) -> Optional[str]:
        return self.__get_cpu('bbox_order')

    @bbox_order.setter
    def bbox_order(self, val: str):
        if val not in ["xy", "yx"]:
            raise ValueError("bbox_order must be either 'xy' or 'yx'")
        self.__set_cpu({'bbox_order': val})
        if self.__intrinsics_has_key('cpu') and len(self.__intrinsics['cpu']) == 0:
            self.__intrinsics.pop('cpu')

    @property
    def softmax(self) -> Optional[bool]:
        return self.__get_cpu('softmax')

    @softmax.setter
    def softmax(self, val: Optional[bool]):
        if val is None:
            return

        if val:
            self.__set_cpu({'softmax': val})
        elif self.__intrinsics_has_key('cpu'):
            self.__intrinsics['cpu'].pop('softmax', None)

        if self.__intrinsics_has_key('cpu') and len(self.__intrinsics['cpu']) == 0:
            self.__intrinsics.pop('cpu')

    @property
    def postprocess(self) -> Optional[str]:
        return self.__get_cpu('post_processing')

    @postprocess.setter
    def postprocess(self, val: str):
        if val != "":
            self.__set_cpu({'post_processing': val})
        elif self.__intrinsics_has_key('cpu'):
            self.__intrinsics['cpu'].pop('post_processing', None)

        if self.__intrinsics_has_key('cpu') and len(self.__intrinsics['cpu']) == 0:
            self.__intrinsics.pop('cpu')

    @property
    def preserve_aspect_ratio(self) -> Optional[bool]:
        if not self.__intrinsics_has_key('input_aspect_ratio'):
            return None
        ar = self.__intrinsics['input_aspect_ratio']
        return ar['width'] == ar['height']

    @preserve_aspect_ratio.setter
    def preserve_aspect_ratio(self, val: Optional[bool]):
        if val is None:
            return

        if val:
            iar = {'input_aspect_ratio': {'width': 1, 'height': 1}}
            self.__top_level_validated_insert(iar)
        elif self.__intrinsics_has_key('input_aspect_ratio'):
            self.__intrinsics.pop('input_aspect_ratio')

    @property
    def labels(self) -> Optional[List[str]]:
        return self.__intrinsics['classes'].get('labels', None) if self.__intrinsics_has_key('classes') else None

    @labels.setter
    def labels(self, val: List[str]):
        if len(val) != 0:
            classes = {'labels': val}
            jsonschema.validate({'classes': classes}, self.__schema | {'additionalProperties': False})

            classes = self.__intrinsics_get_key('classes', {}) | classes
            if self.__intrinsics:
                self.__intrinsics['classes'] = classes
            else:
                self.__intrinsics = {'classes': classes}
        elif self.__intrinsics_has_key('classes'):
            self.__intrinsics['classes'].pop('labels', None)
            if len(self.__intrinsics['classes']) == 0:
                self.__intrinsics.pop('classes')

    @property
    def ignore_dash_labels(self) -> Optional[bool]:
        return self.__intrinsics['classes'].get('ignore_undefined', None) if self.__intrinsics_has_key('classes') else None

    @ignore_dash_labels.setter
    def ignore_dash_labels(self, val: Optional[bool]):
        if val is None:
            return

        if val:
            iu = {'ignore_undefined': val}
            jsonschema.validate({'classes': iu}, self.__schema | {'additionalProperties': False})

            classes = {'classes': self.__intrinsics_get_key('classes', {}) | iu}
            self.__intrinsics = self.__intrinsics | classes if self.__intrinsics else classes
        elif self.__intrinsics_has_key('classes'):
            self.__intrinsics['classes'].pop('ignore_undefined', None)
            if len(self.__intrinsics['classes']) == 0:
                self.__intrinsics.pop('classes')


class IMX500:
    def __init__(self, network_file: str, camera_id: str = ''):
        self.device_fd = None
        self.fw_progress = None
        self.fw_progress_chunk = None
        self.__cfg = {'network_file': network_file, 'input_tensor': {}}

        imx500_device_id = None
        spi_device_id = None
        for i in range(32):
            test_dir = f'/sys/class/video4linux/v4l-subdev{i}/device'
            module_dir = f'{test_dir}/driver/module'
            id_dir = f'{test_dir}/of_node'
            if os.path.exists(module_dir) and os.path.islink(module_dir) and os.path.islink(id_dir) \
                    and 'imx500' in os.readlink(module_dir):
                if camera_id == '' or (camera_id in os.readlink(id_dir)):
                    self.device_fd = open(f'/dev/v4l-subdev{i}', 'rb+', buffering=0)
                    imx500_device_id = os.readlink(test_dir).split('/')[-1]
                    spi_device_id = imx500_device_id.replace('001a', '0040')
                    camera_info = Picamera2.global_camera_info()
                    self.__camera_num = next((c['Num'] for c in camera_info if c['Model'] == 'imx500'
                                              and c['Id'] in os.readlink(id_dir)))
                    break

        if self.device_fd is None:
            raise RuntimeError('IMX500: Requested camera dev-node not found')

        # Progress status specific debugfs entries.
        if imx500_device_id:
            self.fw_progress = open(f'/sys/kernel/debug/imx500-fw:{imx500_device_id}/fw_progress', 'r')
        if spi_device_id:
            self.fw_progress_chunk = open(f'/sys/kernel/debug/rp2040-spi:{spi_device_id}/transfer_progress', 'r')

        if self.config['network_file'] != '':
            self.__set_network_firmware(os.path.abspath(self.config['network_file']))
            self.__ni_from_network(os.path.abspath(self.config['network_file']))

        if 'norm_val' not in self.__cfg['input_tensor']:
            self.__cfg['input_tensor']['norm_val'] = [-2048, -2048, -2048]
        if 'norm_shift' not in self.__cfg:
            self.__cfg['input_tensor']['norm_shift'] = [4, 4, 4]
        if 'div_val' not in self.__cfg:
            self.__cfg['input_tensor']['div_val'] = [1024, 1024, 1024]
        if 'div_shift' not in self.__cfg:
            self.__cfg['input_tensor']['div_shift'] = 6

        full_sensor = self.__get_full_sensor_resolution()
        self.set_inference_roi_abs(full_sensor.to_tuple())

    @staticmethod
    def __get_full_sensor_resolution():
        """Full sensor resolution as a Rectangle object."""
        return Rectangle(0, 0, 4056, 3040)

    def __del__(self):
        if self.device_fd:
            self.device_fd.close()

    @property
    def camera_num(self):
        return self.__camera_num

    @property
    def config(self) -> dict:
        return self.__cfg

    @property
    def network_intrinsics(self) -> Optional[NetworkIntrinsics]:
        return self.__cfg.get('intrinsics', None)

    def convert_inference_coords(self, coords: tuple, metadata: dict, picam2: Picamera2, stream='main') -> tuple:
        """Convert relative inference coordinates into the output image coordinates space."""
        isp_output_size = Size(*picam2.camera_configuration()[stream]['size'])
        sensor_output_size = Size(*picam2.camera_configuration()['raw']['size'])
        scaler_crop = Rectangle(*metadata['ScalerCrop'])

        y0, x0, y1, x1 = coords
        full_sensor = self.__get_full_sensor_resolution()
        width, height = full_sensor.size.to_tuple()
        obj = Rectangle(
            *np.maximum(
                np.array([x0 * width, y0 * height, (x1 - x0) * width, (y1 - y0) * height]),
                0,
            ).astype(np.int32)
        )
        out = self.__get_obj_scaled(obj, isp_output_size, scaler_crop, sensor_output_size)
        return out.to_tuple()

    def get_fw_upload_progress(self, stage_req) -> tuple:
        """Returns the current progress of the fw upload in the form of (current, total)."""
        progress_block = 0
        progress_chunk = 0
        size = 0
        stage = 0

        if self.fw_progress:
            self.fw_progress.seek(0)
            progress = self.fw_progress.readline().strip().split()
            stage = int(progress[0])
            progress_block = int(progress[1])
            size = int(progress[2])

        if self.fw_progress_chunk:
            self.fw_progress_chunk.seek(0)
            progress_chunk = int(self.fw_progress_chunk.readline().strip())

        if stage == stage_req:
            return (min(progress_block + progress_chunk, size), size)
        else:
            return (0, 0)

    def show_network_fw_progress_bar(self):
        p = multiprocessing.Process(target=self.__do_progress_bar,
                                    args=(FW_NETWORK_STAGE, 'Network Firmware Upload'))
        p.start()
        p.join(0)

    def __do_progress_bar(self, stage_req, title):
        with tqdm(unit='bytes', unit_scale=True, unit_divisor=1024, desc=title, leave=True) as t:
            last_update = 0
            while True:
                current, total = self.get_fw_upload_progress(stage_req)
                if total:
                    t.total = total
                    t.update(current - last_update)
                    last_update = current
                    if current > 0.95 * total:
                        t.update(total - last_update)
                        break
                time.sleep(0.5)

    def get_roi_scaled(self, request: CompletedRequest, stream="main") -> tuple:
        """Get the region of interest (ROI) in output image coordinates space."""
        picam2 = request.picam2
        isp_output_size = self.get_isp_output_size(picam2, stream)
        sensor_output_size = self.get_isp_output_size(picam2, 'raw')
        scaler_crop = Rectangle(*request.get_metadata()['ScalerCrop'])
        obj = self.__get_full_sensor_resolution()
        roi = self.__get_obj_scaled(obj, isp_output_size, scaler_crop, sensor_output_size)
        return roi.to_tuple()

    @staticmethod
    def get_isp_output_size(picam2, stream="main") -> tuple:
        return Size(*picam2.camera_configuration()[stream]['size'])

    def __get_obj_scaled(self, obj, isp_output_size, scaler_crop, sensor_output_size) -> Rectangle:
        """Scale the object coordinates based on the camera configuration and sensor properties."""
        full_sensor = self.__get_full_sensor_resolution()
        width, height = full_sensor.size.to_tuple()
        sensor_crop = scaler_crop.scaled_by(sensor_output_size, full_sensor.size)

        # Make sure the object is bound to the user requested ROI.
        if 'roi' in self.config and self.config['roi'] != Rectangle(0, 0, 0, 0):
            obj = obj.bounded_to(self.config['roi'])

        obj_sensor = obj.scaled_by(sensor_output_size, Size(width, height))
        obj_bound = obj_sensor.bounded_to(sensor_crop)
        obj_translated = obj_bound.translated_by(-sensor_crop.topLeft)
        obj_scaled = obj_translated.scaled_by(isp_output_size, sensor_crop.size)
        return obj_scaled

    def get_input_size(self) -> tuple:
        """Get the model input tensor size as (width, height)."""
        return self.config['input_tensor_size']

    def input_tensor_image(self, input_tensor):
        """Convert input tensor in planar format to interleaved RGB."""
        width = self.config['input_tensor']['width']
        height = self.config['input_tensor']['height']
        r1 = np.array(input_tensor, dtype=np.uint8).astype(np.int32).reshape((3,) + (height, width))
        r1 = r1[(2, 1, 0), :, :]
        norm_val = self.config['input_tensor']['norm_val']
        norm_shift = self.config['input_tensor']['norm_shift']
        div_val = self.config['input_tensor']['div_val']
        div_shift = self.config['input_tensor']['div_shift']
        for i in [0, 1, 2]:
            r1[i] = ((((r1[i] << norm_shift[i]) - norm_val[i]) << div_shift) // div_val[i]) & 0xff

        return np.transpose(r1, (1, 2, 0)).astype(np.uint8)

    def get_outputs(self, metadata: dict, add_batch=False) -> Optional[list[np.ndarray]]:
        """Get the model outputs."""
        output_tensor = metadata.get('CnnOutputTensor')
        if not output_tensor:
            return None

        np_output = np.fromiter(output_tensor, dtype=np.float32)
        output_shapes = self.get_output_shapes(metadata)
        offset = 0
        outputs = []
        for tensor_shape in output_shapes:
            size = np.prod(tensor_shape)
            reshaped_tensor = np_output[offset:offset + size].reshape(tensor_shape, order='F')
            if add_batch:
                reshaped_tensor = np.expand_dims(reshaped_tensor, 0)
            outputs.append(reshaped_tensor)
            offset += size
        return outputs

    def get_output_shapes(self, metadata: dict) -> list[tuple[int]]:
        """Get the model output shapes if no output return empty list."""
        output_tensor_info = metadata.get('CnnOutputTensorInfo')
        if not output_tensor_info:
            return []
        output_tensor_info = self.__get_output_tensor_info(output_tensor_info)['info']
        return [o['size'] for o in output_tensor_info]

    def set_inference_roi_abs(self, roi: tuple):
        """
        Set the absolute inference image crop.

        Specify an absolute region of interest in the form a (left, top, width, height) crop for the input inference
        image. The co-ordinates are based on the full sensor resolution.
        """
        roi = Rectangle(*roi)
        roi = roi.bounded_to(self.__get_full_sensor_resolution())

        r = (ctypes.c_uint32 * 4)()
        r[0] = roi.x
        r[1] = roi.y
        r[2] = roi.width
        r[3] = roi.height

        c = (v4l2_ext_control * 1)()
        c[0].p_u32 = r
        c[0].id = ROI_CTRL_ID
        c[0].size = 16

        ctrl = v4l2_ext_controls()
        ctrl.count = 1
        ctrl.controls = c

        try:
            fcntl.ioctl(self.device_fd, VIDIOC_S_EXT_CTRLS, ctrl)
            self.__cfg['roi'] = roi
        except OSError as err:
            print(f'IMX500: Unable to set ROI control in the device driver: {err}')

    def set_inference_aspect_ratio(self, aspect_ratio: tuple):
        """
        Set the aspect ratio of the inference image.

        Specify a pixel aspect ratio needed for the input inference image relative to the full sensor resolution.
        This simply calculates an ROI based on a centre crop and calls set_inference_roi_abs().
        """
        f = self.__get_full_sensor_resolution()
        r = f.size.bounded_to_aspect_ratio(Size(aspect_ratio[0], aspect_ratio[1]))
        r = r.centered_to(f.center).enclosed_in(f)
        self.set_inference_roi_abs(r.to_tuple())

    def set_auto_aspect_ratio(self):
        """Set the inference image crop to presereve the input tensor aspect ratio."""
        self.set_inference_aspect_ratio(self.config['input_tensor_size'])

    def __get_output_tensor_info(self, tensor_info) -> dict:
        """Return the network string along with a list of output tensor parameters."""
        if type(tensor_info) not in [bytes, bytearray]:
            tensor_info = bytes(tensor_info)

        size = ctypes.sizeof(CnnOutputTensorInfoExported)
        if len(tensor_info) != size:
            raise ValueError(f'tensor info length {len(tensor_info)} does not match expected size {size}')

        # Create an instance of the struct and copy data into it
        parsed = CnnOutputTensorInfoExported()
        ctypes.memmove(ctypes.addressof(parsed), tensor_info, size)

        result = {
            'network_name': parsed.network_name.decode('utf-8').strip('\x00'),
            'num_tensors': parsed.num_tensors,
            'info': []
        }

        for t in parsed.info[0:parsed.num_tensors]:
            info = {
                'tensor_data_num': t.tensor_data_num,
                'num_dimensions': t.num_dimensions,
                'size': list(t.size)[0:t.num_dimensions],
            }
            result['info'].append(info)

        return result

    def __get_input_tensor_info(self, tensor_info) -> tuple[str, int, int, int]:
        """Return the input tensor parameters in the form (network_name, width, height, num_channels)."""
        NETWORK_NAME_LEN = 64
        tensor_fmt = f'{NETWORK_NAME_LEN}sIII'

        if type(tensor_info) not in [bytes, bytearray]:
            tensor_info = bytes(tensor_info)

        network_name, width, height, num_channels = struct.unpack(tensor_fmt, tensor_info)
        network_name = network_name.decode('utf-8').rstrip('\0')
        return (network_name, width, height, num_channels)

    @staticmethod
    def get_kpi_info(metadata: dict) -> Optional[tuple[float, float]]:
        """Return the KPI parameters in the form (dnn_runtime, dsp_runtime) in milliseconds."""
        kpi_info = metadata.get('CnnKpiInfo')
        if kpi_info is None:
            return None
        dnn_runtime, dsp_runtime = kpi_info[0], kpi_info[1]
        return dnn_runtime / 1000, dsp_runtime / 1000

    def __set_network_firmware(self, network_filename: str):
        """Provides a firmware rpk file to upload to the IMX500. This must be called before Picamera2 is configured."""
        if not os.path.isfile(network_filename):
            raise RuntimeError(f'Firmware file {network_filename} does not exist.')

        fd = os.open(network_filename, os.O_RDONLY)
        if fd:
            ctrl = v4l2_control()
            ctrl.id = NETWORK_FW_FD_CTRL_ID
            ctrl.value = fd

            try:
                fcntl.ioctl(self.device_fd, VIDIOC_S_CTRL, ctrl)
                print('\n------------------------------------------------------------------------------------------------------------------\n'  # noqa
                      'NOTE: Loading network firmware onto the IMX500 can take several minutes, please do not close down the application.'  # noqa
                      '\n------------------------------------------------------------------------------------------------------------------\n', file=sys.stderr)  # noqa
            except OSError as err:
                raise RuntimeError(f'IMX500: Unable to set network firmware {network_filename}: {err}')
            finally:
                os.close(fd)

    def __ni_from_network(self, network_filename: str):
        """Extracts 'network_info.txt' from CPIO-archive appended to the network rpk."""
        with open(network_filename, 'rb') as fp:
            fw = memoryview(fp.read())

        # Iterate through network firmware discarding blocks
        cpio_offset = 0
        while True:
            # Parse header (+ current block size)
            (magic, size) = struct.unpack('>4sI', fw[:8])
            if not magic == b'9464':
                break
            fw = fw[size + 60:]
            # Ensure footer is as expected
            (magic,) = struct.unpack('4s', fw[:4])
            if not magic == b'3695':
                raise RuntimeError(f'No matching footer found in firmware file {network_filename}')
            fw = fw[4:]
            cpio_offset += size + 64

        cpio_fd = os.open(network_filename, os.O_RDONLY)
        os.lseek(cpio_fd, cpio_offset, os.SEEK_SET)

        with fd_reader(cpio_fd) as archive:
            for entry in archive:
                if 'network_info.txt' == str(entry):
                    self.__cfg['network_info_raw'] = b''.join(entry.get_blocks())
                elif 'network_intrinsics' == str(entry):
                    self.__cfg['intrinsics'] = NetworkIntrinsics(json.loads(b''.join(entry.get_blocks())))

        os.close(cpio_fd)

        if 'network_info_raw' not in self.__cfg:
            return

        res = {}
        buf = io.StringIO(self.__cfg['network_info_raw'].decode('ascii'))
        for line in buf:
            key, value = line.strip().split('=')
            if key == 'networkID':
                nid: int = 0
                for idx, x in enumerate(value):
                    nid |= (ord(x) - ord('0')) << (20 - idx * 4)
                res[key] = nid
            if key == 'apParamSize':
                res[key] = int(value)
            if key == 'networkNum':
                res[key] = int(value)

        res['network'] = {}
        networks = self.__cfg['network_info_raw'].decode('ascii').split('networkOrdinal=')[1:]
        for nw in networks:
            buf = io.StringIO(nw)
            nw_idx = int(buf.readline())
            nw_properties = {}
            for line in buf:
                key, value = line.strip().split('=')
                nw_properties[key] = value
            res['network'][nw_idx] = nw_properties

        if len(res['network']) != res['networkNum']:
            raise RuntimeError('Insufficient networkNum settings in network_info.txt')

        self.__cfg['network_info'] = res

        # Extract some input tensor config params
        self.__cfg['input_tensor']['width'] = int(res['network'][0]['inputTensorWidth'])
        self.__cfg['input_tensor']['height'] = int(res['network'][0]['inputTensorHeight'])
        self.__cfg['input_tensor_size'] = (self.config['input_tensor']['width'],
                                           self.config['input_tensor']['height'])

        input_format = self.__cfg['network_info']['network'][0]['inputTensorFormat']
        inputTensorNorm_K03 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K03'], 0)
        inputTensorNorm_K13 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K13'], 0)
        inputTensorNorm_K23 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K23'], 0)
        inputTensorNorm_K00 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K00'], 0)
        inputTensorNorm_K22 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K22'], 0)
        inputTensorNorm_K02 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K02'], 0)
        inputTensorNorm_K20 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K20'], 0)
        inputTensorNorm_K11 = int(self.__cfg['network_info']['network'][0]['inputTensorNorm_K11'], 0)

        self.__cfg['input_tensor']['input_format'] = input_format

        if input_format == 'RGB' or input_format == 'BGR':
            norm_val_0 = \
                inputTensorNorm_K03 if ((inputTensorNorm_K03 >> 12) & 1) == 0 else -((~inputTensorNorm_K03 + 1) & 0x1fff)
            norm_val_1 = \
                inputTensorNorm_K13 if ((inputTensorNorm_K13 >> 12) & 1) == 0 else -((~inputTensorNorm_K13 + 1) & 0x1fff)
            norm_val_2 = \
                inputTensorNorm_K23 if ((inputTensorNorm_K23 >> 12) & 1) == 0 else -((~inputTensorNorm_K23 + 1) & 0x1fff)
            norm_val = [norm_val_0, norm_val_1, norm_val_2]
            self.__cfg['input_tensor']['norm_val'] = norm_val
            norm_shift = [4, 4, 4]
            self.__cfg['input_tensor']['norm_shift'] = norm_shift
            if input_format == 'RGB':
                div_val_0 = \
                    inputTensorNorm_K00 if ((inputTensorNorm_K00 >> 11) & 1) == 0 else -((~inputTensorNorm_K00 + 1) & 0x0fff)
                div_val_2 =\
                    inputTensorNorm_K22 if ((inputTensorNorm_K22 >> 11) & 1) == 0 else -((~inputTensorNorm_K22 + 1) & 0x0fff)
            else:
                div_val_0 = \
                    inputTensorNorm_K02 if ((inputTensorNorm_K02 >> 11) & 1) == 0 else -((~inputTensorNorm_K02 + 1) & 0x0fff)
                div_val_2 = \
                    inputTensorNorm_K20 if ((inputTensorNorm_K20 >> 11) & 1) == 0 else -((~inputTensorNorm_K20 + 1) & 0x0fff)
            div_val_1 = \
                inputTensorNorm_K11 if ((inputTensorNorm_K11 >> 11) & 1) == 0 else -((~inputTensorNorm_K11 + 1) & 0x0fff)
            self.__cfg['input_tensor']['div_val'] = [div_val_0, div_val_1, div_val_2]
            self.__cfg['input_tensor']['div_shift'] = 6



"""
Yolov5 postprocessing

This code is based on:
https://github.com/ultralytics/ultralytics
"""
from typing import Tuple

import cv2
import numpy as np

from picamera2.devices.imx500.postprocess import (
    BoxFormat, combined_nms, combined_nms_seg,
    convert_to_ymin_xmin_ymax_xmax_format, crop_mask, nms)


def postprocess_yolov8_detection(outputs: Tuple[np.ndarray, np.ndarray],
                                 conf: float = 0.3,
                                 iou_thres: float = 0.7,
                                 max_out_dets: int = 50) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Postprocess the outputs of a YOLOv8 model for object detection

    Args:
        outputs (Tuple[np.ndarray, np.ndarray]): Tuple containing the model outputs for bounding boxes and class predictions.
        conf (float, optional): Confidence threshold for bounding box predictions. Default is 0.3
        iou_thres (float, optional): IoU (Intersection over Union) threshold for Non-Maximum Suppression (NMS). Default is 0.7.
        max_out_dets (int, optional): Maximum number of output detections to keep after NMS. Default is 50.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple containing the post-processed bounding boxes,
            their corresponding scores, and categories.
    """
    feat_sizes = np.array([80, 40, 20])
    stride_sizes = np.array([8, 16, 32])
    a, s = (x.transpose() for x in make_anchors_yolo_v8(feat_sizes, stride_sizes, 0.5))

    y_bb, y_cls = outputs
    dbox = dist2bbox_yolo_v8(y_bb, a, xywh=True, dim=1) * s
    detect_out = np.concatenate((dbox, y_cls), 1)

    xd = detect_out.transpose([0, 2, 1])

    return combined_nms(xd[..., :4], xd[..., 4:84], iou_thres, conf, max_out_dets)


def postprocess_yolov8_keypoints(outputs: Tuple[np.ndarray, np.ndarray, np.ndarray],
                                 conf: float = 0.3,
                                 iou_thres: float = 0.7,
                                 max_out_dets: int = 300) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Postprocess the outputs of a YOLOv8 model for object detection and pose estimation.

    Args:
        outputs (Tuple[np.ndarray, np.ndarray, np.ndarray]): Tuple containing the model outputs for bounding boxes,
        class predictions, and keypoint predictions.
        conf (float, optional): Confidence threshold for bounding box predictions. Default is 0.3
        iou_thres (float, optional): IoU (Intersection over Union) threshold for Non-Maximum Suppression (NMS). Default is 0.7.
        max_out_dets (int, optional): Maximum number of output detections to keep after NMS. Default is 300.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple containing the post-processed bounding boxes, their
        corresponding scores, and keypoints.

    """
    kpt_shape = (17, 3)
    feat_sizes = np.array([80, 40, 20])
    stride_sizes = np.array([8, 16, 32])
    a, s = (x.transpose() for x in make_anchors_yolo_v8(feat_sizes, stride_sizes, 0.5))

    y_bb, y_cls, kpts = outputs
    dbox = dist2bbox_yolo_v8(y_bb, a, xywh=True, dim=1) * s
    detect_out = np.concatenate((dbox, y_cls), 1)
    # additional part for pose estimation
    ndim = kpt_shape[1]
    pred_kpt = kpts.copy()
    if ndim == 3:
        pred_kpt[:, 2::3] = 1 / (1 + np.exp(-pred_kpt[:, 2::3]))  # sigmoid (WARNING: inplace .sigmoid_() Apple MPS bug)
    pred_kpt[:, 0::ndim] = (pred_kpt[:, 0::ndim] * 2.0 + (a[0] - 0.5)) * s
    pred_kpt[:, 1::ndim] = (pred_kpt[:, 1::ndim] * 2.0 + (a[1] - 0.5)) * s

    x = np.concatenate([detect_out.transpose([2, 1, 0]).squeeze(), pred_kpt.transpose([2, 1, 0]).squeeze()], 1)
    x = x[(x[:, 4] > conf)]
    x = x[np.argsort(-x[:, 4])[:8400]]
    x[..., :4] = convert_to_ymin_xmin_ymax_xmax_format(x[..., :4], BoxFormat.XC_YC_W_H)
    boxes = x[..., :4]
    scores = x[..., 4]

    # Original post-processing part
    valid_indexs = nms(boxes, scores, iou_thres=iou_thres, max_out_dets=max_out_dets)
    x = x[valid_indexs]
    nms_bbox = x[:, :4]
    nms_scores = x[:, 4]
    nms_kpts = x[:, 5:]

    return nms_bbox, nms_scores, nms_kpts


def postprocess_yolov8_inst_seg(outputs: Tuple[np.ndarray, np.ndarray, np.ndarray],
                                conf: float = 0.001,
                                iou_thres: float = 0.7,
                                max_out_dets: int = 300) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    feat_sizes = np.array([80, 40, 20])
    stride_sizes = np.array([8, 16, 32])
    a, s = (x.transpose() for x in make_anchors_yolo_v8(feat_sizes, stride_sizes, 0.5))

    y_bb, y_cls, ymask_weights, y_masks = outputs
    dbox = dist2bbox_yolo_v8(y_bb, a, xywh=True, dim=1) * s
    detect_out = np.concatenate((dbox, y_cls), 1)

    xd = detect_out.transpose([0, 2, 1])
    nms_bbox, nms_scores, nms_classes, ymask_weights = combined_nms_seg(xd[..., :4], xd[..., 4:84],
                                                                        ymask_weights, iou_thres, conf, max_out_dets)[0]
    if len(nms_scores) == 0:
        final_masks = y_masks
    else:
        y_masks = y_masks.squeeze(0)
        ymask_weights = ymask_weights.transpose(1, 0)
        final_masks = np.tensordot(ymask_weights, y_masks, axes=([0], [0]))

    return nms_bbox, nms_scores, nms_classes, final_masks


def make_anchors_yolo_v8(feats, strides, grid_cell_offset=0.5):
    """Generate anchors from features."""
    anchor_points, stride_tensor = [], []
    assert feats is not None
    for i, stride in enumerate(strides):
        h, w = feats[i], feats[i]
        sx = np.arange(stop=w) + grid_cell_offset  # shift x
        sy = np.arange(stop=h) + grid_cell_offset  # shift y
        sy, sx = np.meshgrid(sy, sx, indexing='ij')
        anchor_points.append(np.stack((sx, sy), -1).reshape((-1, 2)))
        stride_tensor.append(np.full((h * w, 1), stride))
    return np.concatenate(anchor_points), np.concatenate(stride_tensor)


def dist2bbox_yolo_v8(distance, anchor_points, xywh=True, dim=-1):
    """Transform distance(ltrb) to box(xywh or xyxy)."""
    lt, rb = np.split(distance, 2, axis=dim)
    x1y1 = anchor_points - lt
    x2y2 = anchor_points + rb
    if xywh:
        c_xy = (x1y1 + x2y2) / 2
        wh = x2y2 - x1y1
        return np.concatenate((c_xy, wh), dim)  # xywh bbox
    return np.concatenate((x1y1, x2y2), dim)  # xyxy bbox


def pad_with_zeros(mask, roi, isp_output_size):
    new_shape = (isp_output_size.width, isp_output_size.height, mask.shape[2])
    padded_mask = np.zeros(new_shape, dtype=mask.dtype)
    padded_mask[roi.x:roi.x + mask.shape[0], roi.y:roi.y + mask.shape[1], :] = mask
    return padded_mask


def process_masks(masks, boxes, roi, isp_output_size):
    # Crop masks based on bounding boxes
    masks = crop_mask(masks, boxes)

    # Apply sigmoid function to normalize masks
    masks = 1 / (1 + np.exp(-masks))
    masks = np.transpose(masks, (2, 1, 0))  # Change to HWC format

    # Resize masks to model input size
    masks = cv2.resize(masks, (roi.height, roi.width), interpolation=cv2.INTER_LINEAR)

    # Ensure masks are in the correct shape
    masks = np.expand_dims(masks, -1) if len(masks.shape) == 2 else masks

    masks = pad_with_zeros(masks, roi, isp_output_size)

    # Ensure masks are in the correct shape
    masks = np.expand_dims(masks, -1) if len(masks.shape) == 2 else masks
    masks = np.transpose(masks, (2, 1, 0))  # Change back to CHW format
    return masks
